import unittest
import canopy
from munch import Munch


class DictToObjectTest(unittest.TestCase):

    def test_it_should_perform_deep_conversion(self):
        data = {
            'a': 1,
            'b': {
                'c': 2
            },
            'd': 3
        }

        output = canopy.dict_to_object(data)

        self.assertTrue(isinstance(output.b, Munch))
        self.assertEqual(output.b.c, 2)

    def test_it_should_not_perform_conversion_twice(self):
        data = {
            'a': 1,
            'b': {
                'c': 2
            },
            'd': 3
        }

        output = canopy.dict_to_object(data)
        output2 = canopy.dict_to_object(output)

        self.assertIs(output, output2)
        self.assertIs(output.b, output2.b)




import unittest
import canopy


class GetStudyTypeDefinitionForSimVersionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.study_type_definition = canopy.openapi.StudyTypeDefinition(
            'dynamicLap',
            'Dynamic Lap',
            ['DynamicLap'],
            ['car', 'track', 'weather'],
            valid_for_transient=False,
            valid_for_inline=False,
            previous_definitions=[
                canopy.openapi.IPreviousDefinitionStudyTypeDefinition(
                    sim_version='1.50',
                    definition=canopy.openapi.StudyTypeDefinition(
                        'dynamicLap',
                        'Dynamic Lap 50',
                        ['DynamicLap'],
                        ['car', 'track'],
                        valid_for_transient=False,
                        valid_for_inline=False,
                        previous_definitions=[])),
                canopy.openapi.IPreviousDefinitionStudyTypeDefinition(
                    sim_version='1.100',
                    definition=canopy.openapi.StudyTypeDefinition(
                        'dynamicLap',
                        'Dynamic Lap 100',
                        ['DynamicLap'],
                        ['car', 'track'],
                        valid_for_transient=False,
                        valid_for_inline=False,
                        previous_definitions=[])),
            ])

    def test_it_should_return_root_definition_when_no_sim_version(self):
        self.assertEqual(
            canopy.get_study_type_definition_for_sim_version(self.study_type_definition),
            self.study_type_definition)

    def test_it_should_return_old_version_when_required(self):
        self.assertEqual(
            canopy.get_study_type_definition_for_sim_version(self.study_type_definition, '1.84'),
            self.study_type_definition.previous_definitions[1].definition)
        self.assertEqual(
            canopy.get_study_type_definition_for_sim_version(self.study_type_definition, '1.50'),
            self.study_type_definition.previous_definitions[0].definition)

    def test_it_should_return_latest_version_when_required(self):
        self.assertEqual(
            canopy.get_study_type_definition_for_sim_version(self.study_type_definition, '1.123'),
            self.study_type_definition)



# coding: utf-8

# flake8: noqa
"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from canopy.openapi.models.account_settings_put_request import AccountSettingsPutRequest
from canopy.openapi.models.additional_tests import AdditionalTests
from canopy.openapi.models.admin_tenant_settings import AdminTenantSettings
from canopy.openapi.models.admin_tenant_settings_builder import AdminTenantSettingsBuilder
from canopy.openapi.models.admin_tenant_settings_pool_settings import AdminTenantSettingsPoolSettings
from canopy.openapi.models.allocation_state import AllocationState
from canopy.openapi.models.auto_scale_run import AutoScaleRun
from canopy.openapi.models.auto_scale_run_error import AutoScaleRunError
from canopy.openapi.models.availability_result import AvailabilityResult
from canopy.openapi.models.availability_result_additional_tests import AvailabilityResultAdditionalTests
from canopy.openapi.models.batch_create_configs_result import BatchCreateConfigsResult
from canopy.openapi.models.blob_access_information import BlobAccessInformation
from canopy.openapi.models.canopy_document import CanopyDocument
from canopy.openapi.models.canopy_document_support_session import CanopyDocumentSupportSession
from canopy.openapi.models.channel_import_mapping import ChannelImportMapping
from canopy.openapi.models.channel_settings import ChannelSettings
from canopy.openapi.models.chart_settings import ChartSettings
from canopy.openapi.models.collated_label_definitions import CollatedLabelDefinitions
from canopy.openapi.models.collated_label_definitions_worksheet import CollatedLabelDefinitionsWorksheet
from canopy.openapi.models.collated_worksheet_labels import CollatedWorksheetLabels
from canopy.openapi.models.collated_worksheet_labels_label_definitions import CollatedWorksheetLabelsLabelDefinitions
from canopy.openapi.models.compute_node_result import ComputeNodeResult
from canopy.openapi.models.compute_node_state import ComputeNodeState
from canopy.openapi.models.condition_operator import ConditionOperator
from canopy.openapi.models.config_column_label_definitions import ConfigColumnLabelDefinitions
from canopy.openapi.models.config_decrypt_with_metadata_request import ConfigDecryptWithMetadataRequest
from canopy.openapi.models.config_encrypt_with_metadata_request import ConfigEncryptWithMetadataRequest
from canopy.openapi.models.config_hash import ConfigHash
from canopy.openapi.models.config_owner_data import ConfigOwnerData
from canopy.openapi.models.config_post_config_request import ConfigPostConfigRequest
from canopy.openapi.models.config_post_configs_request import ConfigPostConfigsRequest
from canopy.openapi.models.config_put_config_owner_request import ConfigPutConfigOwnerRequest
from canopy.openapi.models.config_put_config_request import ConfigPutConfigRequest
from canopy.openapi.models.config_reference import ConfigReference
from canopy.openapi.models.config_reference_default import ConfigReferenceDefault
from canopy.openapi.models.config_reference_tenant import ConfigReferenceTenant
from canopy.openapi.models.config_resolved_labels import ConfigResolvedLabels
from canopy.openapi.models.config_resolved_labels_reference import ConfigResolvedLabelsReference
from canopy.openapi.models.config_resolved_reference import ConfigResolvedReference
from canopy.openapi.models.config_resolved_reference_data import ConfigResolvedReferenceData
from canopy.openapi.models.config_type_definition import ConfigTypeDefinition
from canopy.openapi.models.config_type_metadata import ConfigTypeMetadata
from canopy.openapi.models.config_type_state import ConfigTypeState
from canopy.openapi.models.config_upgrade_config_deprecated_request import ConfigUpgradeConfigDeprecatedRequest
from canopy.openapi.models.data_to_decrypt import DataToDecrypt
from canopy.openapi.models.data_to_encrypt import DataToEncrypt
from canopy.openapi.models.default_config_id import DefaultConfigId
from canopy.openapi.models.default_config_reference import DefaultConfigReference
from canopy.openapi.models.default_custom_property_names import DefaultCustomPropertyNames
from canopy.openapi.models.document_custom_property_data import DocumentCustomPropertyData
from canopy.openapi.models.document_custom_property_group import DocumentCustomPropertyGroup
from canopy.openapi.models.document_group_result import DocumentGroupResult
from canopy.openapi.models.document_name_result import DocumentNameResult
from canopy.openapi.models.document_type_custom_property_groups import DocumentTypeCustomPropertyGroups
from canopy.openapi.models.document_user_information import DocumentUserInformation
from canopy.openapi.models.documents_and_continuation_token import DocumentsAndContinuationToken
from canopy.openapi.models.duplicate_configs_data import DuplicateConfigsData
from canopy.openapi.models.duplicate_configs_result import DuplicateConfigsResult
from canopy.openapi.models.email_confirmation_data import EmailConfirmationData
from canopy.openapi.models.file_download_metadata import FileDownloadMetadata
from canopy.openapi.models.get_account_settings_result import GetAccountSettingsResult
from canopy.openapi.models.get_admin_tenant_settings_query_result import GetAdminTenantSettingsQueryResult
from canopy.openapi.models.get_admin_tenant_settings_query_result_settings import GetAdminTenantSettingsQueryResultSettings
from canopy.openapi.models.get_all_support_sessions_query_result import GetAllSupportSessionsQueryResult
from canopy.openapi.models.get_all_support_sessions_query_result_user_information import GetAllSupportSessionsQueryResultUserInformation
from canopy.openapi.models.get_all_tenants_study_statistics_query_result import GetAllTenantsStudyStatisticsQueryResult
from canopy.openapi.models.get_config_names_query_result import GetConfigNamesQueryResult
from canopy.openapi.models.get_config_query_result import GetConfigQueryResult
from canopy.openapi.models.get_config_query_result_config import GetConfigQueryResultConfig
from canopy.openapi.models.get_config_versions_query_result import GetConfigVersionsQueryResult
from canopy.openapi.models.get_configs_query_result import GetConfigsQueryResult
from canopy.openapi.models.get_configs_query_result_query_results import GetConfigsQueryResultQueryResults
from canopy.openapi.models.get_decrypted_data_query_result import GetDecryptedDataQueryResult
from canopy.openapi.models.get_encrypted_data_query_result import GetEncryptedDataQueryResult
from canopy.openapi.models.get_pool_status_query_result import GetPoolStatusQueryResult
from canopy.openapi.models.get_pools_item import GetPoolsItem
from canopy.openapi.models.get_pools_item_interval import GetPoolsItemInterval
from canopy.openapi.models.get_pools_query_result import GetPoolsQueryResult
from canopy.openapi.models.get_sim_version_document_query_result import GetSimVersionDocumentQueryResult
from canopy.openapi.models.get_sim_version_document_query_result_document import GetSimVersionDocumentQueryResultDocument
from canopy.openapi.models.get_sim_version_documents_query_result import GetSimVersionDocumentsQueryResult
from canopy.openapi.models.get_sim_version_downloads_query_result import GetSimVersionDownloadsQueryResult
from canopy.openapi.models.get_studies_query_result import GetStudiesQueryResult
from canopy.openapi.models.get_studies_query_result_query_results import GetStudiesQueryResultQueryResults
from canopy.openapi.models.get_study_download_url_query_result import GetStudyDownloadUrlQueryResult
from canopy.openapi.models.get_study_job_metadata_query_result import GetStudyJobMetadataQueryResult
from canopy.openapi.models.get_study_job_metadata_query_result_access_information import GetStudyJobMetadataQueryResultAccessInformation
from canopy.openapi.models.get_study_job_query_result import GetStudyJobQueryResult
from canopy.openapi.models.get_study_jobs_query_result import GetStudyJobsQueryResult
from canopy.openapi.models.get_study_jobs_query_result_query_results import GetStudyJobsQueryResultQueryResults
from canopy.openapi.models.get_study_query_result import GetStudyQueryResult
from canopy.openapi.models.get_study_query_result_access_information import GetStudyQueryResultAccessInformation
from canopy.openapi.models.get_study_types_query_result import GetStudyTypesQueryResult
from canopy.openapi.models.get_support_session_query_result import GetSupportSessionQueryResult
from canopy.openapi.models.get_tenant_access_information_query_result import GetTenantAccessInformationQueryResult
from canopy.openapi.models.get_tenant_billable_stored_simulation_count_query_result import GetTenantBillableStoredSimulationCountQueryResult
from canopy.openapi.models.get_tenant_channel_import_mappings_query_result import GetTenantChannelImportMappingsQueryResult
from canopy.openapi.models.get_tenant_channel_whitelists_query_result import GetTenantChannelWhitelistsQueryResult
from canopy.openapi.models.get_tenant_channel_whitelists_query_result_settings import GetTenantChannelWhitelistsQueryResultSettings
from canopy.openapi.models.get_tenant_default_custom_property_names_query_result import GetTenantDefaultCustomPropertyNamesQueryResult
from canopy.openapi.models.get_tenant_default_custom_property_names_query_result_settings import GetTenantDefaultCustomPropertyNamesQueryResultSettings
from canopy.openapi.models.get_tenant_query_result import GetTenantQueryResult
from canopy.openapi.models.get_tenant_settings_sim_version_query_result import GetTenantSettingsSimVersionQueryResult
from canopy.openapi.models.get_tenant_study_statistics_query_result import GetTenantStudyStatisticsQueryResult
from canopy.openapi.models.get_tenant_users_query_result import GetTenantUsersQueryResult
from canopy.openapi.models.get_tenant_users_query_result_user_item import GetTenantUsersQueryResultUserItem
from canopy.openapi.models.get_tenant_worksheet_label_definitions_query_result import GetTenantWorksheetLabelDefinitionsQueryResult
from canopy.openapi.models.get_tenant_worksheet_label_definitions_query_result_label_definitions import GetTenantWorksheetLabelDefinitionsQueryResultLabelDefinitions
from canopy.openapi.models.get_tenants_query_result import GetTenantsQueryResult
from canopy.openapi.models.get_tenants_query_result_tenant_item import GetTenantsQueryResultTenantItem
from canopy.openapi.models.get_upvoty_token_query_result import GetUpvotyTokenQueryResult
from canopy.openapi.models.get_user_roles_query_result import GetUserRolesQueryResult
from canopy.openapi.models.get_user_settings_query_result import GetUserSettingsQueryResult
from canopy.openapi.models.get_user_settings_query_result_settings import GetUserSettingsQueryResultSettings
from canopy.openapi.models.get_wiki_document_query_result import GetWikiDocumentQueryResult
from canopy.openapi.models.get_worksheet_query_result import GetWorksheetQueryResult
from canopy.openapi.models.get_worksheet_query_result_worksheet import GetWorksheetQueryResultWorksheet
from canopy.openapi.models.get_zendesk_token_query_result import GetZendeskTokenQueryResult
from canopy.openapi.models.grant_type_handler_response import GrantTypeHandlerResponse
from canopy.openapi.models.group_operator import GroupOperator
from canopy.openapi.models.i_previous_definition_sim_type_definition import IPreviousDefinitionSimTypeDefinition
from canopy.openapi.models.i_previous_definition_sim_type_definition_definition import IPreviousDefinitionSimTypeDefinitionDefinition
from canopy.openapi.models.i_previous_definition_study_type_definition import IPreviousDefinitionStudyTypeDefinition
from canopy.openapi.models.i_previous_definition_study_type_definition_definition import IPreviousDefinitionStudyTypeDefinitionDefinition
from canopy.openapi.models.identified_user_data import IdentifiedUserData
from canopy.openapi.models.label_definition import LabelDefinition
from canopy.openapi.models.label_definitions import LabelDefinitions
from canopy.openapi.models.list_filter import ListFilter
from canopy.openapi.models.list_filter_condition import ListFilterCondition
from canopy.openapi.models.list_filter_group import ListFilterGroup
from canopy.openapi.models.list_filter_query import ListFilterQuery
from canopy.openapi.models.membership_post_email_confirmation_request import MembershipPostEmailConfirmationRequest
from canopy.openapi.models.membership_post_identified_user_request import MembershipPostIdentifiedUserRequest
from canopy.openapi.models.membership_post_password_reset_confirmation_request import MembershipPostPasswordResetConfirmationRequest
from canopy.openapi.models.membership_post_password_reset_request_request import MembershipPostPasswordResetRequestRequest
from canopy.openapi.models.membership_post_registration_request import MembershipPostRegistrationRequest
from canopy.openapi.models.membership_put_user_role_request import MembershipPutUserRoleRequest
from canopy.openapi.models.name_value_pair import NameValuePair
from canopy.openapi.models.new_batch_config_data import NewBatchConfigData
from canopy.openapi.models.new_config_data import NewConfigData
from canopy.openapi.models.new_sim_version_data import NewSimVersionData
from canopy.openapi.models.new_study_data import NewStudyData
from canopy.openapi.models.new_study_data_source import NewStudyDataSource
from canopy.openapi.models.new_tenant_data import NewTenantData
from canopy.openapi.models.new_worksheet_data import NewWorksheetData
from canopy.openapi.models.new_worksheet_data_outline import NewWorksheetDataOutline
from canopy.openapi.models.order_by_property import OrderByProperty
from canopy.openapi.models.password_reset_confirmation_data import PasswordResetConfirmationData
from canopy.openapi.models.password_reset_request_data import PasswordResetRequestData
from canopy.openapi.models.pool_settings import PoolSettings
from canopy.openapi.models.pool_state import PoolState
from canopy.openapi.models.pool_type import PoolType
from canopy.openapi.models.post_study_result import PostStudyResult
from canopy.openapi.models.registration_data import RegistrationData
from canopy.openapi.models.resolved_label import ResolvedLabel
from canopy.openapi.models.resolved_label_numeric_statistics import ResolvedLabelNumericStatistics
from canopy.openapi.models.resolved_statistic_label import ResolvedStatisticLabel
from canopy.openapi.models.sim_type_definition import SimTypeDefinition
from canopy.openapi.models.sim_type_definition_input_telemetry_channels import SimTypeDefinitionInputTelemetryChannels
from canopy.openapi.models.sim_type_input_telemetry_channel import SimTypeInputTelemetryChannel
from canopy.openapi.models.sim_type_input_telemetry_channel_evaluation_definition import SimTypeInputTelemetryChannelEvaluationDefinition
from canopy.openapi.models.sim_type_input_telemetry_channels import SimTypeInputTelemetryChannels
from canopy.openapi.models.sim_type_input_telemetry_evaluated_channel import SimTypeInputTelemetryEvaluatedChannel
from canopy.openapi.models.sim_type_input_telemetry_evaluated_channel_suffix import SimTypeInputTelemetryEvaluatedChannelSuffix
from canopy.openapi.models.sim_type_state import SimTypeState
from canopy.openapi.models.sim_version_document_name_result import SimVersionDocumentNameResult
from canopy.openapi.models.sim_version_post_sim_version_request import SimVersionPostSimVersionRequest
from canopy.openapi.models.simulation_column_label_definitions import SimulationColumnLabelDefinitions
from canopy.openapi.models.simulation_input import SimulationInput
from canopy.openapi.models.simulation_resolved_labels import SimulationResolvedLabels
from canopy.openapi.models.study_blob_access_information import StudyBlobAccessInformation
from canopy.openapi.models.study_document import StudyDocument
from canopy.openapi.models.study_document_study_document_data_source import StudyDocumentStudyDocumentDataSource
from canopy.openapi.models.study_documents_and_continuation_token import StudyDocumentsAndContinuationToken
from canopy.openapi.models.study_input_hash import StudyInputHash
from canopy.openapi.models.study_input_hashes import StudyInputHashes
from canopy.openapi.models.study_job_documents_and_continuation_token import StudyJobDocumentsAndContinuationToken
from canopy.openapi.models.study_post_study_request import StudyPostStudyRequest
from canopy.openapi.models.study_put_study_request import StudyPutStudyRequest
from canopy.openapi.models.study_reference import StudyReference
from canopy.openapi.models.study_resolved_labels import StudyResolvedLabels
from canopy.openapi.models.study_resolved_labels_reference import StudyResolvedLabelsReference
from canopy.openapi.models.study_resolved_reference import StudyResolvedReference
from canopy.openapi.models.study_resolved_reference_data import StudyResolvedReferenceData
from canopy.openapi.models.study_resolved_reference_data_study_document import StudyResolvedReferenceDataStudyDocument
from canopy.openapi.models.study_state import StudyState
from canopy.openapi.models.study_type_definition import StudyTypeDefinition
from canopy.openapi.models.study_type_state import StudyTypeState
from canopy.openapi.models.support_session import SupportSession
from canopy.openapi.models.support_session_data import SupportSessionData
from canopy.openapi.models.support_session_put_support_session_request import SupportSessionPutSupportSessionRequest
from canopy.openapi.models.support_session_response import SupportSessionResponse
from canopy.openapi.models.support_sessions_request_metadata import SupportSessionsRequestMetadata
from canopy.openapi.models.tenancy_post_tenant_request import TenancyPostTenantRequest
from canopy.openapi.models.tenancy_put_tenant_request import TenancyPutTenantRequest
from canopy.openapi.models.tenant_channel_whitelists import TenantChannelWhitelists
from canopy.openapi.models.tenant_config_reference import TenantConfigReference
from canopy.openapi.models.tenant_default_custom_property_names import TenantDefaultCustomPropertyNames
from canopy.openapi.models.tenant_information import TenantInformation
from canopy.openapi.models.tenant_settings_put_admin_tenant_settings_request import TenantSettingsPutAdminTenantSettingsRequest
from canopy.openapi.models.tenant_settings_put_tenant_channel_import_mappings_request import TenantSettingsPutTenantChannelImportMappingsRequest
from canopy.openapi.models.tenant_settings_put_tenant_channel_whitelists_request import TenantSettingsPutTenantChannelWhitelistsRequest
from canopy.openapi.models.tenant_settings_put_tenant_default_custom_property_names_request import TenantSettingsPutTenantDefaultCustomPropertyNamesRequest
from canopy.openapi.models.tenant_settings_put_tenant_settings_sim_version_request import TenantSettingsPutTenantSettingsSimVersionRequest
from canopy.openapi.models.tenant_settings_put_tenant_worksheet_label_definitions_request import TenantSettingsPutTenantWorksheetLabelDefinitionsRequest
from canopy.openapi.models.tenant_statistics import TenantStatistics
from canopy.openapi.models.test_auto_scale_formula_query_result import TestAutoScaleFormulaQueryResult
from canopy.openapi.models.test_auto_scale_formula_query_result_auto_scale_run import TestAutoScaleFormulaQueryResultAutoScaleRun
from canopy.openapi.models.text_document import TextDocument
from canopy.openapi.models.text_document_optional_content import TextDocumentOptionalContent
from canopy.openapi.models.updated_account_settings import UpdatedAccountSettings
from canopy.openapi.models.updated_admin_tenant_settings import UpdatedAdminTenantSettings
from canopy.openapi.models.updated_admin_tenant_settings_settings import UpdatedAdminTenantSettingsSettings
from canopy.openapi.models.updated_channel_import_mappings import UpdatedChannelImportMappings
from canopy.openapi.models.updated_config_data import UpdatedConfigData
from canopy.openapi.models.updated_study_data import UpdatedStudyData
from canopy.openapi.models.updated_tenant_channel_whitelists import UpdatedTenantChannelWhitelists
from canopy.openapi.models.updated_tenant_data import UpdatedTenantData
from canopy.openapi.models.updated_tenant_default_custom_property_names import UpdatedTenantDefaultCustomPropertyNames
from canopy.openapi.models.updated_tenant_settings_sim_version import UpdatedTenantSettingsSimVersion
from canopy.openapi.models.updated_user_settings import UpdatedUserSettings
from canopy.openapi.models.updated_user_settings_settings import UpdatedUserSettingsSettings
from canopy.openapi.models.updated_worksheet_data import UpdatedWorksheetData
from canopy.openapi.models.updated_worksheet_label_definitions import UpdatedWorksheetLabelDefinitions
from canopy.openapi.models.upgrade_config_data import UpgradeConfigData
from canopy.openapi.models.upgrade_config_query_result import UpgradeConfigQueryResult
from canopy.openapi.models.user_information import UserInformation
from canopy.openapi.models.user_role_data import UserRoleData
from canopy.openapi.models.user_settings import UserSettings
from canopy.openapi.models.user_settings_builder import UserSettingsBuilder
from canopy.openapi.models.user_settings_put_user_settings_request import UserSettingsPutUserSettingsRequest
from canopy.openapi.models.versioned_document_metadata import VersionedDocumentMetadata
from canopy.openapi.models.worksheet import Worksheet
from canopy.openapi.models.worksheet_config import WorksheetConfig
from canopy.openapi.models.worksheet_config_reference import WorksheetConfigReference
from canopy.openapi.models.worksheet_outline import WorksheetOutline
from canopy.openapi.models.worksheet_post_duplicate_configs_request import WorksheetPostDuplicateConfigsRequest
from canopy.openapi.models.worksheet_post_worksheet_request import WorksheetPostWorksheetRequest
from canopy.openapi.models.worksheet_put_worksheet_request import WorksheetPutWorksheetRequest
from canopy.openapi.models.worksheet_resolved_labels import WorksheetResolvedLabels
from canopy.openapi.models.worksheet_resolved_references import WorksheetResolvedReferences
from canopy.openapi.models.worksheet_row import WorksheetRow
from canopy.openapi.models.worksheet_row_study import WorksheetRowStudy
from canopy.openapi.models.worksheet_study import WorksheetStudy
from canopy.openapi.models.worksheet_study_reference import WorksheetStudyReference



from asyncio import Future
from typing import List, Optional

import asyncio
import pytest
import numpy as np

import canopy
import integration_tests

default_car_name = 'Canopy F1 Car 2019'
test_car_name = 'Python Automated Test Car'
test_car_name_2 = 'Python Automated Test Car 2'
test_car_property_key = 'python-automated-test'
test_car_property_value = 'true'
test_car_custom_properties = {
    test_car_property_key: test_car_property_value,
}


class State(object):
    configs_to_remove: List[str] = []
    default_car: Optional[canopy.LocalConfig] = None
    test_car_config_id: Optional[str] = None
    test_car_config_id_2: Optional[str] = None


@pytest.mark.asyncio
class TestConfig:

    @pytest.fixture(autouse=True, scope='class')
    def state(self) -> State:
        return State()

    async def test_0100_it_should_load_a_default_config(self, state: State):
        async with integration_tests.Environment() as environment:
            state.default_car = await canopy.load_default_config(
                environment.session,
                'car',
                default_car_name)

            assert state.default_car is not None
            assert default_car_name == state.default_car.name

    async def test_0200_it_should_create_a_config(self, state: State):

        async with integration_tests.Environment() as environment:
            state.test_car_config_id = await canopy.create_config(
                environment.session,
                'car',
                test_car_name,
                state.default_car.raw_data,
                test_car_custom_properties)

            assert state.test_car_config_id is not None

            state.configs_to_remove.append(state.test_car_config_id)

    async def test_0210_it_should_create_a_config_with_no_custom_properties(self, state: State):

        async with integration_tests.Environment() as environment:
            state.test_car_config_id_2 = await canopy.create_config(
                environment.session,
                'car',
                test_car_name_2,
                state.default_car.raw_data)

            assert state.test_car_config_id_2 is not None

            state.configs_to_remove.append(state.test_car_config_id_2)

    async def test_0220_it_should_create_a_config_containing_numpy_arrays(self, state: State):
        async with integration_tests.Environment() as environment:
            state.default_car.data.chassis.rUndertrayFront = np.array(state.default_car.data.chassis.rUndertrayFront)

            state.test_car_config_id = await canopy.create_config(
                environment.session,
                'car',
                test_car_name,
                state.default_car.raw_data,
                test_car_custom_properties)

            assert state.test_car_config_id is not None

            state.configs_to_remove.append(state.test_car_config_id)

    async def test_0300_it_should_find_a_config_by_name(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.find_config(
                environment.session,
                'car',
                owner_username=environment.data.authentication.username,
                name=test_car_name)

            assert car.document.document_id == state.test_car_config_id

    async def test_0400_it_should_find_a_config_by_custom_properties(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.find_config(
                environment.session,
                'car',
                owner_username=environment.data.authentication.username,
                custom_properties=test_car_custom_properties)

            assert car.document.document_id == state.test_car_config_id

    async def test_0500_it_should_find_a_config_by_multiple_criteria(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.find_config(
                environment.session,
                'car',
                owner_username=environment.data.authentication.username,
                name=test_car_name,
                custom_properties=test_car_custom_properties)

            assert car.document.document_id == state.test_car_config_id

    async def test_0600_it_should_load_a_config_by_id(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.load_config(
                environment.session,
                state.test_car_config_id)

            assert car.document.document_id == state.test_car_config_id

    async def test_0700_it_should_convert_configs_to_local_configs(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.load_config(
                environment.session,
                state.test_car_config_id)
            local_car = car.to_local_config()

            assert car.document.document_id == state.test_car_config_id
            assert local_car.config_id == state.test_car_config_id

    async def test_0700_it_should_convert_configs_to_local_configs_with_no_custom_properties(self, state: State):
        async with integration_tests.Environment() as environment:
            car = await canopy.load_config(
                environment.session,
                state.test_car_config_id_2)
            local_car = car.to_local_config()

            assert car.document.document_id == state.test_car_config_id_2
            assert local_car.config_id == state.test_car_config_id_2

    async def test_9000_it_should_remove_configs(self, state: State):
        async with integration_tests.Environment() as environment:
            tasks: List[Future] = []
            for config_id in state.configs_to_remove:
                tasks.append(asyncio.ensure_future(canopy.delete_config(
                    environment.session,
                    config_id)))
            await asyncio.gather(*tasks)



from threading import Thread
from typing import Optional, Union
from aiohttp.client_exceptions import ClientResponseError, ClientConnectionError, ServerTimeoutError, ClientError

import numpy as np
import pandas as pd
import canopy
import logging
import asyncio

logger = logging.getLogger(__name__)


async def load_channel(
        session: canopy.Session,
        job_access_information: canopy.openapi.BlobAccessInformation,
        sim_type: str,
        channel_name: str,
        vector_metadata: Optional[pd.DataFrame] = None,
        semaphore: Optional[asyncio.Semaphore] = None) -> Optional[canopy.LoadedChannel]:
    sim_type = canopy.ensure_sim_type_string(sim_type)

    if semaphore is None:
        semaphore = asyncio.Semaphore(1)

    async with semaphore:
        if vector_metadata is None:
            vector_metadata = await canopy.load_vector_metadata(session, job_access_information, sim_type)

        if vector_metadata is None:
            return None

        if channel_name not in vector_metadata.index:
            logger.debug('Channel not found: %s', channel_name)
            return None

        channel_metadata = vector_metadata.xs(channel_name)

        points_count: int = channel_metadata['NPtsInChannel']
        units: str = channel_metadata['units']

        file_name = ''.join([sim_type, '_', channel_name, '.bin'])
        channel_url = ''.join([job_access_information.url, file_name, job_access_information.access_signature])

        channel_bytes: Optional[bytes] = await session.try_load_bytes(
            channel_url,
            f'"{file_name}" from "{job_access_information.url}"')

        if channel_bytes is None:
            return None

        if points_count * 4 == len(channel_bytes):
            data_type = np.float32
        else:
            data_type = np.float64
        channel_data: np.array = np.frombuffer(channel_bytes, data_type)

        loaded_channel = canopy.LoadedChannel(channel_name, units, channel_data)
        return loaded_channel


async def read_channel_data(channel_url: str, session: canopy.Session) -> bytes:
    async with session.async_client_session.get(
            channel_url,
            raise_for_status=True,
            proxy=session.configuration.proxy) as response:
        return await response.read()



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class ListFilter(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'items_per_page': 'int',
        'continuation_token': 'str',
        'order_by_custom_property': 'object',
        'order_by_property': 'OrderByProperty',
        'order_by_descending': 'bool',
        'filter_name': 'str',
        'query': 'ListFilterQuery',
        'include_if_delete_requested': 'bool',
        'include_if_has_parent_worksheet': 'bool'
    }

    attribute_map = {
        'items_per_page': 'itemsPerPage',
        'continuation_token': 'continuationToken',
        'order_by_custom_property': 'orderByCustomProperty',
        'order_by_property': 'orderByProperty',
        'order_by_descending': 'orderByDescending',
        'filter_name': 'filterName',
        'query': 'query',
        'include_if_delete_requested': 'includeIfDeleteRequested',
        'include_if_has_parent_worksheet': 'includeIfHasParentWorksheet'
    }

    def __init__(self, items_per_page=None, continuation_token=None, order_by_custom_property=None, order_by_property=None, order_by_descending=None, filter_name=None, query=None, include_if_delete_requested=None, include_if_has_parent_worksheet=None, local_vars_configuration=None):  # noqa: E501
        """ListFilter - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._items_per_page = None
        self._continuation_token = None
        self._order_by_custom_property = None
        self._order_by_property = None
        self._order_by_descending = None
        self._filter_name = None
        self._query = None
        self._include_if_delete_requested = None
        self._include_if_has_parent_worksheet = None
        self.discriminator = None

        if items_per_page is not None:
            self.items_per_page = items_per_page
        self.continuation_token = continuation_token
        self.order_by_custom_property = order_by_custom_property
        self.order_by_property = order_by_property
        if order_by_descending is not None:
            self.order_by_descending = order_by_descending
        self.filter_name = filter_name
        self.query = query
        if include_if_delete_requested is not None:
            self.include_if_delete_requested = include_if_delete_requested
        if include_if_has_parent_worksheet is not None:
            self.include_if_has_parent_worksheet = include_if_has_parent_worksheet

    @property
    def items_per_page(self):
        """Gets the items_per_page of this ListFilter.  # noqa: E501


        :return: The items_per_page of this ListFilter.  # noqa: E501
        :rtype: int
        """
        return self._items_per_page

    @items_per_page.setter
    def items_per_page(self, items_per_page):
        """Sets the items_per_page of this ListFilter.


        :param items_per_page: The items_per_page of this ListFilter.  # noqa: E501
        :type items_per_page: int
        """

        self._items_per_page = items_per_page

    @property
    def continuation_token(self):
        """Gets the continuation_token of this ListFilter.  # noqa: E501


        :return: The continuation_token of this ListFilter.  # noqa: E501
        :rtype: str
        """
        return self._continuation_token

    @continuation_token.setter
    def continuation_token(self, continuation_token):
        """Sets the continuation_token of this ListFilter.


        :param continuation_token: The continuation_token of this ListFilter.  # noqa: E501
        :type continuation_token: str
        """

        self._continuation_token = continuation_token

    @property
    def order_by_custom_property(self):
        """Gets the order_by_custom_property of this ListFilter.  # noqa: E501


        :return: The order_by_custom_property of this ListFilter.  # noqa: E501
        :rtype: object
        """
        return self._order_by_custom_property

    @order_by_custom_property.setter
    def order_by_custom_property(self, order_by_custom_property):
        """Sets the order_by_custom_property of this ListFilter.


        :param order_by_custom_property: The order_by_custom_property of this ListFilter.  # noqa: E501
        :type order_by_custom_property: object
        """

        self._order_by_custom_property = order_by_custom_property

    @property
    def order_by_property(self):
        """Gets the order_by_property of this ListFilter.  # noqa: E501


        :return: The order_by_property of this ListFilter.  # noqa: E501
        :rtype: OrderByProperty
        """
        return self._order_by_property

    @order_by_property.setter
    def order_by_property(self, order_by_property):
        """Sets the order_by_property of this ListFilter.


        :param order_by_property: The order_by_property of this ListFilter.  # noqa: E501
        :type order_by_property: OrderByProperty
        """

        self._order_by_property = order_by_property

    @property
    def order_by_descending(self):
        """Gets the order_by_descending of this ListFilter.  # noqa: E501


        :return: The order_by_descending of this ListFilter.  # noqa: E501
        :rtype: bool
        """
        return self._order_by_descending

    @order_by_descending.setter
    def order_by_descending(self, order_by_descending):
        """Sets the order_by_descending of this ListFilter.


        :param order_by_descending: The order_by_descending of this ListFilter.  # noqa: E501
        :type order_by_descending: bool
        """

        self._order_by_descending = order_by_descending

    @property
    def filter_name(self):
        """Gets the filter_name of this ListFilter.  # noqa: E501


        :return: The filter_name of this ListFilter.  # noqa: E501
        :rtype: str
        """
        return self._filter_name

    @filter_name.setter
    def filter_name(self, filter_name):
        """Sets the filter_name of this ListFilter.


        :param filter_name: The filter_name of this ListFilter.  # noqa: E501
        :type filter_name: str
        """

        self._filter_name = filter_name

    @property
    def query(self):
        """Gets the query of this ListFilter.  # noqa: E501


        :return: The query of this ListFilter.  # noqa: E501
        :rtype: ListFilterQuery
        """
        return self._query

    @query.setter
    def query(self, query):
        """Sets the query of this ListFilter.


        :param query: The query of this ListFilter.  # noqa: E501
        :type query: ListFilterQuery
        """

        self._query = query

    @property
    def include_if_delete_requested(self):
        """Gets the include_if_delete_requested of this ListFilter.  # noqa: E501


        :return: The include_if_delete_requested of this ListFilter.  # noqa: E501
        :rtype: bool
        """
        return self._include_if_delete_requested

    @include_if_delete_requested.setter
    def include_if_delete_requested(self, include_if_delete_requested):
        """Sets the include_if_delete_requested of this ListFilter.


        :param include_if_delete_requested: The include_if_delete_requested of this ListFilter.  # noqa: E501
        :type include_if_delete_requested: bool
        """

        self._include_if_delete_requested = include_if_delete_requested

    @property
    def include_if_has_parent_worksheet(self):
        """Gets the include_if_has_parent_worksheet of this ListFilter.  # noqa: E501


        :return: The include_if_has_parent_worksheet of this ListFilter.  # noqa: E501
        :rtype: bool
        """
        return self._include_if_has_parent_worksheet

    @include_if_has_parent_worksheet.setter
    def include_if_has_parent_worksheet(self, include_if_has_parent_worksheet):
        """Sets the include_if_has_parent_worksheet of this ListFilter.


        :param include_if_has_parent_worksheet: The include_if_has_parent_worksheet of this ListFilter.  # noqa: E501
        :type include_if_has_parent_worksheet: bool
        """

        self._include_if_has_parent_worksheet = include_if_has_parent_worksheet

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ListFilter):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ListFilter):
            return True

        return self.to_dict() != other.to_dict()



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from canopy.openapi.api.account_settings_api import AccountSettingsApi
from canopy.openapi.api.availability_api import AvailabilityApi
from canopy.openapi.api.config_api import ConfigApi
from canopy.openapi.api.end_to_end_test_inbox_api import EndToEndTestInboxApi
from canopy.openapi.api.list_filter_api import ListFilterApi
from canopy.openapi.api.membership_api import MembershipApi
from canopy.openapi.api.pool_api import PoolApi
from canopy.openapi.api.sim_version_api import SimVersionApi
from canopy.openapi.api.study_api import StudyApi
from canopy.openapi.api.support_session_api import SupportSessionApi
from canopy.openapi.api.tenancy_api import TenancyApi
from canopy.openapi.api.tenant_settings_api import TenantSettingsApi
from canopy.openapi.api.token_api import TokenApi
from canopy.openapi.api.user_settings_api import UserSettingsApi
from canopy.openapi.api.worksheet_api import WorksheetApi



import unittest
import canopy


class SimVersionToNumberTest(unittest.TestCase):

    def test_it_should_convert_to_number(self):
        self.assertEqual(canopy.sim_version_to_number('1.1234'), 1234)

    def test_it_can_be_used_to_compare_sim_versions(self):
        self.assertTrue(canopy.sim_version_to_number('1.1') == canopy.sim_version_to_number('1.1'))
        self.assertTrue(canopy.sim_version_to_number('1.1') < canopy.sim_version_to_number('1.2'))
        self.assertTrue(canopy.sim_version_to_number('1.2') < canopy.sim_version_to_number('1.10'))
        self.assertTrue(canopy.sim_version_to_number('1.10') < canopy.sim_version_to_number('1.100'))
        self.assertTrue(canopy.sim_version_to_number('1.20') < canopy.sim_version_to_number('1.1020'))




from typing import Any, Dict, Optional

import canopy
import json


class LocalConfig(object):
    _converted_data: Optional[Any] = None

    def __init__(
            self,
            config_type: str,
            name: str,
            data: Any,
            properties: Optional[Dict[str, str]] = None,
            notes: Optional[str] = None,
            user_id: Optional[str] = None,
            config_id: Optional[str] = None,
            is_edited: bool = False):

        self._config_type = config_type
        self.name = name
        self._data = data
        self.properties = properties
        self.notes = notes
        self.user_id = user_id
        self.config_id = config_id
        self.is_edited = is_edited

    @property
    def config_type(self) -> str:
        return self._config_type

    @property
    def data(self) -> Any:
        if self._converted_data is None:
            if self._config_type == canopy.Constants.config_sub_tree_document_type:
                self._converted_data = canopy.dict_to_object(self._data['definition'])
            else:
                self._converted_data = canopy.dict_to_object(self._data)
        return self._converted_data

    @property
    def is_data_converted(self) -> bool:
        return self._converted_data is not None

    @property
    def raw_data(self) -> Any:
        return self._converted_data if self.is_data_converted else self._data

    def to_dict(self):
        return {
            'config_type': self._config_type,
            'name': self.name,
            'properties': self.properties,
            'notes': self.notes,
            'user_id': self.user_id,
            'config_id': self.config_id,
            'is_edited': self.is_edited,
            'is_data_converted': self.is_data_converted,
            'data': self.raw_data
        }

    def __repr__(self):
        return 'canopy.LocalConfig(%r,%r,%r,properties=%r,notes=%r,user_id=%r,config_id=%r,is_edited=%r)' % \
               (
                   self._config_type,
                   self.name,
                   self._converted_data.__dict__ if self.is_data_converted else self._data,
                   self.properties,
                   self.notes,
                   self.user_id,
                   self.config_id,
                   self.is_edited
               )

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2, default=str)



from typing import Optional, Dict
from urllib3 import make_headers


class ProxyConfiguration(object):
    def __init__(
            self,
            url: str,
            username: Optional[str] = None,
            password: Optional[str] = None):

        self._url = url
        self._username = username
        self._password = password

        if '://' not in url:
            raise RuntimeError('Proxy URL must contain scheme such as "http://" or "https://".')

        headers = None
        auth_url = url
        if username is not None and password is not None:
            auth_string = f'{username}:{password}'
            headers = make_headers(proxy_basic_auth=auth_string)
            auth_url = url.replace('://', f'://{auth_string}@')

        self._headers = headers
        self._auth_url = auth_url

    @property
    def url(self) -> str:
        return self._url

    @property
    def username(self) -> Optional[str]:
        return self._username

    @property
    def password(self) -> Optional[str]:
        return self._password

    @property
    def auth_url(self) -> str:
        return self._auth_url

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class SupportSession(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'is_open': 'bool',
        'modified_date': 'datetime',
        'modified_tenant_id': 'object',
        'modified_user_id': 'object',
        'responses': 'list[SupportSessionResponse]'
    }

    attribute_map = {
        'is_open': 'isOpen',
        'modified_date': 'modifiedDate',
        'modified_tenant_id': 'modifiedTenantId',
        'modified_user_id': 'modifiedUserId',
        'responses': 'responses'
    }

    def __init__(self, is_open=None, modified_date=None, modified_tenant_id=None, modified_user_id=None, responses=None, local_vars_configuration=None):  # noqa: E501
        """SupportSession - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._is_open = None
        self._modified_date = None
        self._modified_tenant_id = None
        self._modified_user_id = None
        self._responses = None
        self.discriminator = None

        self.is_open = is_open
        self.modified_date = modified_date
        self.modified_tenant_id = modified_tenant_id
        self.modified_user_id = modified_user_id
        self.responses = responses

    @property
    def is_open(self):
        """Gets the is_open of this SupportSession.  # noqa: E501


        :return: The is_open of this SupportSession.  # noqa: E501
        :rtype: bool
        """
        return self._is_open

    @is_open.setter
    def is_open(self, is_open):
        """Sets the is_open of this SupportSession.


        :param is_open: The is_open of this SupportSession.  # noqa: E501
        :type is_open: bool
        """
        if self.local_vars_configuration.client_side_validation and is_open is None:  # noqa: E501
            raise ValueError("Invalid value for `is_open`, must not be `None`")  # noqa: E501

        self._is_open = is_open

    @property
    def modified_date(self):
        """Gets the modified_date of this SupportSession.  # noqa: E501


        :return: The modified_date of this SupportSession.  # noqa: E501
        :rtype: datetime
        """
        return self._modified_date

    @modified_date.setter
    def modified_date(self, modified_date):
        """Sets the modified_date of this SupportSession.


        :param modified_date: The modified_date of this SupportSession.  # noqa: E501
        :type modified_date: datetime
        """
        if self.local_vars_configuration.client_side_validation and modified_date is None:  # noqa: E501
            raise ValueError("Invalid value for `modified_date`, must not be `None`")  # noqa: E501

        self._modified_date = modified_date

    @property
    def modified_tenant_id(self):
        """Gets the modified_tenant_id of this SupportSession.  # noqa: E501


        :return: The modified_tenant_id of this SupportSession.  # noqa: E501
        :rtype: object
        """
        return self._modified_tenant_id

    @modified_tenant_id.setter
    def modified_tenant_id(self, modified_tenant_id):
        """Sets the modified_tenant_id of this SupportSession.


        :param modified_tenant_id: The modified_tenant_id of this SupportSession.  # noqa: E501
        :type modified_tenant_id: object
        """
        if self.local_vars_configuration.client_side_validation and modified_tenant_id is None:  # noqa: E501
            raise ValueError("Invalid value for `modified_tenant_id`, must not be `None`")  # noqa: E501

        self._modified_tenant_id = modified_tenant_id

    @property
    def modified_user_id(self):
        """Gets the modified_user_id of this SupportSession.  # noqa: E501


        :return: The modified_user_id of this SupportSession.  # noqa: E501
        :rtype: object
        """
        return self._modified_user_id

    @modified_user_id.setter
    def modified_user_id(self, modified_user_id):
        """Sets the modified_user_id of this SupportSession.


        :param modified_user_id: The modified_user_id of this SupportSession.  # noqa: E501
        :type modified_user_id: object
        """
        if self.local_vars_configuration.client_side_validation and modified_user_id is None:  # noqa: E501
            raise ValueError("Invalid value for `modified_user_id`, must not be `None`")  # noqa: E501

        self._modified_user_id = modified_user_id

    @property
    def responses(self):
        """Gets the responses of this SupportSession.  # noqa: E501


        :return: The responses of this SupportSession.  # noqa: E501
        :rtype: list[SupportSessionResponse]
        """
        return self._responses

    @responses.setter
    def responses(self, responses):
        """Sets the responses of this SupportSession.


        :param responses: The responses of this SupportSession.  # noqa: E501
        :type responses: list[SupportSessionResponse]
        """

        self._responses = responses

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SupportSession):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SupportSession):
            return True

        return self.to_dict() != other.to_dict()



from typing import Dict, Union, Optional


def ensure_dict(value: Optional[Union[Dict, object]]) -> Optional[Dict]:
    if value is None:
        return value

    if isinstance(value, dict):
        return value

    return vars(value)



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class GetTenantsQueryResultTenantItem(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'tenant_id': 'object',
        'name': 'object',
        'short_name': 'object',
        'creation_date': 'datetime',
        'is_enabled': 'bool',
        'database_id': 'object',
        'region': 'object'
    }

    attribute_map = {
        'tenant_id': 'tenantId',
        'name': 'name',
        'short_name': 'shortName',
        'creation_date': 'creationDate',
        'is_enabled': 'isEnabled',
        'database_id': 'databaseId',
        'region': 'region'
    }

    def __init__(self, tenant_id=None, name=None, short_name=None, creation_date=None, is_enabled=None, database_id=None, region=None, local_vars_configuration=None):  # noqa: E501
        """GetTenantsQueryResultTenantItem - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._tenant_id = None
        self._name = None
        self._short_name = None
        self._creation_date = None
        self._is_enabled = None
        self._database_id = None
        self._region = None
        self.discriminator = None

        self.tenant_id = tenant_id
        self.name = name
        self.short_name = short_name
        self.creation_date = creation_date
        self.is_enabled = is_enabled
        self.database_id = database_id
        self.region = region

    @property
    def tenant_id(self):
        """Gets the tenant_id of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The tenant_id of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: object
        """
        return self._tenant_id

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        """Sets the tenant_id of this GetTenantsQueryResultTenantItem.


        :param tenant_id: The tenant_id of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type tenant_id: object
        """
        if self.local_vars_configuration.client_side_validation and tenant_id is None:  # noqa: E501
            raise ValueError("Invalid value for `tenant_id`, must not be `None`")  # noqa: E501

        self._tenant_id = tenant_id

    @property
    def name(self):
        """Gets the name of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The name of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: object
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this GetTenantsQueryResultTenantItem.


        :param name: The name of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type name: object
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def short_name(self):
        """Gets the short_name of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The short_name of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: object
        """
        return self._short_name

    @short_name.setter
    def short_name(self, short_name):
        """Sets the short_name of this GetTenantsQueryResultTenantItem.


        :param short_name: The short_name of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type short_name: object
        """
        if self.local_vars_configuration.client_side_validation and short_name is None:  # noqa: E501
            raise ValueError("Invalid value for `short_name`, must not be `None`")  # noqa: E501

        self._short_name = short_name

    @property
    def creation_date(self):
        """Gets the creation_date of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The creation_date of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: datetime
        """
        return self._creation_date

    @creation_date.setter
    def creation_date(self, creation_date):
        """Sets the creation_date of this GetTenantsQueryResultTenantItem.


        :param creation_date: The creation_date of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type creation_date: datetime
        """
        if self.local_vars_configuration.client_side_validation and creation_date is None:  # noqa: E501
            raise ValueError("Invalid value for `creation_date`, must not be `None`")  # noqa: E501

        self._creation_date = creation_date

    @property
    def is_enabled(self):
        """Gets the is_enabled of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The is_enabled of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: bool
        """
        return self._is_enabled

    @is_enabled.setter
    def is_enabled(self, is_enabled):
        """Sets the is_enabled of this GetTenantsQueryResultTenantItem.


        :param is_enabled: The is_enabled of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type is_enabled: bool
        """
        if self.local_vars_configuration.client_side_validation and is_enabled is None:  # noqa: E501
            raise ValueError("Invalid value for `is_enabled`, must not be `None`")  # noqa: E501

        self._is_enabled = is_enabled

    @property
    def database_id(self):
        """Gets the database_id of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The database_id of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: object
        """
        return self._database_id

    @database_id.setter
    def database_id(self, database_id):
        """Sets the database_id of this GetTenantsQueryResultTenantItem.


        :param database_id: The database_id of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type database_id: object
        """
        if self.local_vars_configuration.client_side_validation and database_id is None:  # noqa: E501
            raise ValueError("Invalid value for `database_id`, must not be `None`")  # noqa: E501

        self._database_id = database_id

    @property
    def region(self):
        """Gets the region of this GetTenantsQueryResultTenantItem.  # noqa: E501


        :return: The region of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :rtype: object
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this GetTenantsQueryResultTenantItem.


        :param region: The region of this GetTenantsQueryResultTenantItem.  # noqa: E501
        :type region: object
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GetTenantsQueryResultTenantItem):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GetTenantsQueryResultTenantItem):
            return True

        return self.to_dict() != other.to_dict()



from typing import Mapping, Optional, Dict

import canopy
import pandas as pd
import numpy as np
import json


class StudyJobResult:
    def __init__(
            self,
            session: canopy.Session,
            job_result: Optional[canopy.openapi.GetStudyJobQueryResult],
            job: canopy.openapi.CanopyDocument,
            vector_metadata: Optional[pd.DataFrame],
            vector_data: pd.DataFrame,
            vector_units: Mapping[str, str],
            scalar_data: Mapping[str, float],
            scalar_units: Mapping[str, str],
            inputs: Optional[Dict[str, Dict]]):
        self._session = session
        self._job_result = job_result
        self._job = job
        self._vector_metadata = vector_metadata
        self._vector_data = vector_data
        self._vector_units = vector_units
        self._inputs = canopy.DynamicDictToObject(inputs) if inputs is not None else None
        self._scalar_data = scalar_data
        self._scalar_units = scalar_units

    @property
    def result(self) -> canopy.openapi.GetStudyJobQueryResult:
        return self._job_result

    @property
    def document(self) -> canopy.openapi.CanopyDocument:
        return self._job

    @property
    def vector_metadata(self) -> Optional[pd.DataFrame]:
        return self._vector_metadata

    @property
    def vector_data(self) -> pd.DataFrame:
        return self._vector_data

    @property
    def vector_units(self) -> Mapping[str, str]:
        return self._vector_units

    @property
    def scalar_data(self) -> Mapping[str, float]:
        return self._scalar_data

    @property
    def scalar_units(self) -> Mapping[str, str]:
        return self._scalar_units

    @property
    def inputs(self) -> Optional[canopy.DynamicDictToObject]:
        return self._inputs

    def vector_as(self, channel_name: str, target_units: Optional[str] = None, always_return_copy: bool = False) -> Optional[pd.Series]:
        if target_units is None:
            target_units = self._session.user_settings.get_channel_units(channel_name)

        result: Optional[pd.Series] = None

        if self._vector_data is not None and channel_name in self._vector_data.columns:
            result = self._vector_data[channel_name]

        if result is None:
            return None

        if target_units is None:
            return result.copy() if always_return_copy else result

        source_units = ''
        if channel_name in self._vector_units:
            source_units = self._vector_units[channel_name]

        return self._session.units.convert_series_between_units(
            result,
            source_units,
            target_units,
            inplace=False,
            always_return_copy=always_return_copy)

    def scalar_as(self, channel_name: str, target_unit: Optional[str] = None) -> float:
        if target_unit is None:
            target_unit = self._session.user_settings.get_channel_units(channel_name)

        if channel_name not in self._scalar_data:
            return np.NaN

        result = self._scalar_data[channel_name]

        if target_unit is None:
            return result

        source_unit = ''
        if channel_name in self._scalar_units:
            source_unit = self._scalar_units[channel_name]

        return self._session.units.convert_value_between_units(
            result,
            source_unit,
            target_unit)

    def to_dict(self):
        return {
            'job_result': self._job_result,
            'job': self._job
        }

    def __repr__(self):
        return 'canopy.StudyJobResult(%r,%r)' % \
               (self._job_result, self._job)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2, default=str)


import canopy
import unittest
import pandas as pd
import numpy as np


class ProcessDataFrameTest(unittest.TestCase):

    def test_it_should_do_nothing(self):
        data_frame = pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9],
        })

        copy = data_frame.copy()
        canopy.process_data_frame(data_frame)
        self.assertTrue(data_frame.equals(copy))

    def test_it_should_set_index(self):
        data_frame = pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9],
        })

        canopy.process_data_frame(data_frame, index_column_name='b')
        self.assertTrue(np.array_equal(np.array(data_frame.index), np.array([4, 5, 6])))

    def test_it_should_standardise_no_units_and_allow_empty_strings_for_units(self):
        data_frame = pd.DataFrame({
            'a': [1, 2, 3],
            'units': ['m/s', '()', np.nan],
        })

        canopy.process_data_frame(data_frame)

        self.assertTrue(
            data_frame.equals(
                pd.DataFrame({
                    'a': [1, 2, 3],
                    'units': ['m/s', '', ''],
                })))

    def test_it_should_allow_empty_strings(self):
        data_frame = pd.DataFrame({
            'a': ['1', np.nan, '3'],
            'b': ['4', np.nan, np.nan],
            'c': [np.nan, '8', '9'],
        })

        canopy.process_data_frame(data_frame, empty_string_valid_column_names=['a', 'b'])

        self.assertTrue(
            data_frame.equals(
                pd.DataFrame({
                    'a': ['1', '', '3'],
                    'b': ['4', '', ''],
                    'c': [np.nan, '8', '9'],
                })))



from typing import Optional

import canopy


def get_study_type_definition_for_sim_version(
        item: canopy.openapi.StudyTypeDefinition,
        sim_version: Optional[str] = None) -> canopy.openapi.StudyTypeDefinition:

    if item.previous_definitions is not None and sim_version is not None:
        numeric_sim_version = canopy.sim_version_to_number(sim_version)

        for previous_study_type in item.previous_definitions:
            previous_numeric_sim_version = canopy.sim_version_to_number(previous_study_type.sim_version)
            if previous_numeric_sim_version >= numeric_sim_version:
                return previous_study_type.definition

    return item



from typing import Dict, Any, Optional

import canopy


class DynamicDictToObject:
    def __init__(self, data: Dict[str, Dict]):
        self._data = data

    def __getattr__(self, name: str):
        result: Optional[Any] = None
        if name in self._data:
            result = canopy.dict_to_object(self._data[name])

        setattr(self, name, result)
        return result



from typing import Optional

import canopy
import pandas as pd
import logging

logger = logging.getLogger(__name__)


async def load_vector_metadata(
        session: canopy.Session,
        job_access_information: canopy.openapi.BlobAccessInformation,
        sim_type: str) -> Optional[pd.DataFrame]:

    sim_type = canopy.ensure_sim_type_string(sim_type)

    file_name = ''.join([sim_type, '_VectorMetadata.csv'])
    url = ''.join([job_access_information.url, file_name, job_access_information.access_signature])

    return await canopy.try_load_csv_from_url(
        session,
        url,
        file_name,
        index_column_name='name',
        empty_string_valid_column_names=['units', 'description'])



# coding: utf-8

# flake8: noqa

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "1.0.0"



from canopy.openapi_asyncio.api_client import ApiClient



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class SupportSessionsRequestMetadata(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'database_id': 'str',
        'tenant_id': 'object',
        'query_time_milliseconds': 'float',
        'time_to_query_start_milliseconds': 'float'
    }

    attribute_map = {
        'database_id': 'databaseId',
        'tenant_id': 'tenantId',
        'query_time_milliseconds': 'queryTimeMilliseconds',
        'time_to_query_start_milliseconds': 'timeToQueryStartMilliseconds'
    }

    def __init__(self, database_id=None, tenant_id=None, query_time_milliseconds=None, time_to_query_start_milliseconds=None, local_vars_configuration=None):  # noqa: E501
        """SupportSessionsRequestMetadata - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._database_id = None
        self._tenant_id = None
        self._query_time_milliseconds = None
        self._time_to_query_start_milliseconds = None
        self.discriminator = None

        self.database_id = database_id
        self.tenant_id = tenant_id
        self.query_time_milliseconds = query_time_milliseconds
        self.time_to_query_start_milliseconds = time_to_query_start_milliseconds

    @property
    def database_id(self):
        """Gets the database_id of this SupportSessionsRequestMetadata.  # noqa: E501


        :return: The database_id of this SupportSessionsRequestMetadata.  # noqa: E501
        :rtype: str
        """
        return self._database_id

    @database_id.setter
    def database_id(self, database_id):
        """Sets the database_id of this SupportSessionsRequestMetadata.


        :param database_id: The database_id of this SupportSessionsRequestMetadata.  # noqa: E501
        :type database_id: str
        """
        if self.local_vars_configuration.client_side_validation and database_id is None:  # noqa: E501
            raise ValueError("Invalid value for `database_id`, must not be `None`")  # noqa: E501

        self._database_id = database_id

    @property
    def tenant_id(self):
        """Gets the tenant_id of this SupportSessionsRequestMetadata.  # noqa: E501


        :return: The tenant_id of this SupportSessionsRequestMetadata.  # noqa: E501
        :rtype: object
        """
        return self._tenant_id

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        """Sets the tenant_id of this SupportSessionsRequestMetadata.


        :param tenant_id: The tenant_id of this SupportSessionsRequestMetadata.  # noqa: E501
        :type tenant_id: object
        """
        if self.local_vars_configuration.client_side_validation and tenant_id is None:  # noqa: E501
            raise ValueError("Invalid value for `tenant_id`, must not be `None`")  # noqa: E501

        self._tenant_id = tenant_id

    @property
    def query_time_milliseconds(self):
        """Gets the query_time_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501


        :return: The query_time_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501
        :rtype: float
        """
        return self._query_time_milliseconds

    @query_time_milliseconds.setter
    def query_time_milliseconds(self, query_time_milliseconds):
        """Sets the query_time_milliseconds of this SupportSessionsRequestMetadata.


        :param query_time_milliseconds: The query_time_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501
        :type query_time_milliseconds: float
        """
        if self.local_vars_configuration.client_side_validation and query_time_milliseconds is None:  # noqa: E501
            raise ValueError("Invalid value for `query_time_milliseconds`, must not be `None`")  # noqa: E501

        self._query_time_milliseconds = query_time_milliseconds

    @property
    def time_to_query_start_milliseconds(self):
        """Gets the time_to_query_start_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501


        :return: The time_to_query_start_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501
        :rtype: float
        """
        return self._time_to_query_start_milliseconds

    @time_to_query_start_milliseconds.setter
    def time_to_query_start_milliseconds(self, time_to_query_start_milliseconds):
        """Sets the time_to_query_start_milliseconds of this SupportSessionsRequestMetadata.


        :param time_to_query_start_milliseconds: The time_to_query_start_milliseconds of this SupportSessionsRequestMetadata.  # noqa: E501
        :type time_to_query_start_milliseconds: float
        """
        if self.local_vars_configuration.client_side_validation and time_to_query_start_milliseconds is None:  # noqa: E501
            raise ValueError("Invalid value for `time_to_query_start_milliseconds`, must not be `None`")  # noqa: E501

        self._time_to_query_start_milliseconds = time_to_query_start_milliseconds

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SupportSessionsRequestMetadata):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SupportSessionsRequestMetadata):
            return True

        return self.to_dict() != other.to_dict()



import numpy as np


class LoadedChannel:
    def __init__(self, name: str, units: str, data: np.array):
        self._name = name
        self._units = units
        self._data = data

    @property
    def name(self) -> str:
        return self._name

    @property
    def units(self) -> str:
        return self._units

    @property
    def data(self) -> np.array:
        return self._data



import canopy
import asyncio
import integration_tests

default_car_name = 'Canopy F1 Car 2019'


class TestSynchronous:

    def test_synchronous_call(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        with integration_tests.Environment() as environment:
            default_car = canopy.run(canopy.load_default_config(
                environment.session,
                'car',
                default_car_name))

            assert default_car is not None
            assert default_car_name == default_car.name



from typing import Optional

import canopy
import json


async def load_default_config(
        session: canopy.Session,
        config_type: str,
        name: str,
        sim_version: Optional[str] = None) -> canopy.LocalConfig:

    session.authentication.authenticate()

    document_path = canopy.get_default_config_path(session, config_type, name)

    if sim_version is None:
        sim_version = session.tenant_sim_version.get()

    api = canopy.openapi.SimVersionApi(session.async_client)
    document_result: canopy.openapi.GetSimVersionDocumentQueryResult = await api.sim_version_get_document(sim_version, document_path)

    config_data = json.loads(document_result.document.content)

    return canopy.LocalConfig(
        config_type,
        name,
        config_data)



from typing import Optional, Sequence, Dict

import canopy
import pandas as pd
import logging
import asyncio

logger = logging.getLogger(__name__)


async def load_study_scalar_results(
        session: canopy.Session,
        study_access_information: canopy.openapi.StudyBlobAccessInformation) -> canopy.StudyScalarResults:

    inputs_task = asyncio.ensure_future(
        _load_file(session, 'scalar-inputs.csv', None, None, study_access_information))

    inputs_metadata_task = asyncio.ensure_future(
        _load_file(session, 'scalar-inputs-metadata.csv', 'inputName', ['units', 'description'], study_access_information))

    results_task = asyncio.ensure_future(
        _load_file(session, 'scalar-results.csv', None, None, study_access_information))

    results_metadata_task = asyncio.ensure_future(
        _load_file(session, 'scalar-metadata.csv', None, ['units', 'description'], study_access_information))

    inputs = await inputs_task
    inputs_metadata = await inputs_metadata_task
    results = await results_task
    results_metadata = await results_metadata_task

    if results_metadata is not None:
        results_metadata['fullName'] = results_metadata['name'] + ':' + results_metadata['simType']
        results_metadata.set_index(['fullName'], inplace=True)

    units: Dict[str, str] = {}
    if inputs_metadata is not None:
        for input_name in inputs_metadata.index:
            units[input_name] = inputs_metadata.loc[input_name]['units']
    if results_metadata is not None:
        for input_name in results_metadata.index:
            units[input_name] = results_metadata.loc[input_name]['units']

    return canopy.StudyScalarResults(
        inputs,
        inputs_metadata,
        results,
        results_metadata,
        units)


async def _load_file(
        session: canopy.Session,
        file_name: str,
        index_column_name: Optional[str],
        empty_string_valid_column_names: Optional[Sequence[str]],
        study_access_information: canopy.openapi.StudyBlobAccessInformation) -> Optional[pd.DataFrame]:

    url = ''.join([study_access_information.url, file_name, study_access_information.access_signature])
    return await canopy.try_load_csv_from_url(
        session,
        url,
        file_name,
        index_column_name=index_column_name,
        empty_string_valid_column_names=empty_string_valid_column_names)




from typing import Optional, Dict

import canopy


class TenantUsersCache(object):

    def __init__(self, client: canopy.openapi.ApiClient, authentication: canopy.Authentication):
        self._client = client
        self._authentication = authentication
        self._data: Dict[str, canopy.TenantUsers] = {}

    def reload(self, tenant_id: Optional[str] = None):
        if tenant_id is None:
            tenant_id = self._authentication.tenant_id

        api = canopy.openapi.TenancyApi(self._client)
        result: canopy.openapi.GetTenantUsersQueryResult = api.tenancy_get_tenant_users(tenant_id)
        self._data[tenant_id] = canopy.TenantUsers(result)

    def get(self, tenant_id: Optional[str] = None) -> canopy.TenantUsers:
        if tenant_id is None:
            tenant_id = self._authentication.tenant_id

        if tenant_id not in self._data:
            self.reload(tenant_id)

        if tenant_id not in self._data:
            raise RuntimeError('Users for tenant ID not found.')

        return self._data[tenant_id]



import unittest
import canopy
import pandas as pd
import numpy as np

temperatureK = 373.15
temperatureC = 100.0
temperatureF = 212.0
temperatureK2 = 473.15
temperatureC2 = 200.0
temperatureF2 = 392.0


class UnitsTest(unittest.TestCase):
    def setUp(self):
        self.units = canopy.Units()

    # Specific units.
    def test_specific_units(self):
        self.assertAlmostEqual(
            self.units.convert_value_between_units(1, 'inHg', 'Pa'),
            3386.39, delta=0.01);
        self.assertAlmostEqual(
            self.units.convert_value_between_units(1000, 'Pa', 'inHg'),
            0.2953, delta=0.0001);

    def test_try_get_conversion_to_si(self):
        existing = self.units.try_get_conversion_to_si('F')
        self.assertEqual(existing.factor, 5/9)
        self.assertEqual(existing.offset, 459.67 * 5 / 9)

        missing = self.units.try_get_conversion_to_si('blah')
        self.assertEqual(missing, None)

    def test_get_conversion_to_si(self):
        existing = self.units.get_conversion_to_si('F')
        self.assertEqual(existing.factor, 5/9)
        self.assertEqual(existing.offset, 459.67 * 5 / 9)

        with self.assertRaises(KeyError):
            self.units.get_conversion_to_si('blah')

    def test_get_conversion_to_si_or_default(self):
        existing = self.units.get_conversion_to_si_or_default('F')
        self.assertEqual(existing.factor, 5/9)
        self.assertEqual(existing.offset, 459.67 * 5 / 9)

        missing = self.units.get_conversion_to_si_or_default('blah')
        self.assertEqual(missing.factor, 1)
        self.assertEqual(missing.offset, 0)

    # FROM SI
    def test_convert_value_from_si(self):
        self.assertAlmostEqual(self.units.convert_value_from_si(temperatureK, 'C'), temperatureC, delta=0.01)
        self.assertAlmostEqual(self.units.convert_value_from_si(temperatureK, 'F'), temperatureF, delta=0.01)

        self.assertAlmostEqual(self.units.convert_value_from_si(temperatureK, 'C', True), temperatureK, delta=0.01)
        self.assertAlmostEqual(self.units.convert_value_from_si(temperatureK, 'F', True), temperatureK * 9 / 5, delta=0.01)

        self.assertEqual(self.units.convert_value_from_si(temperatureK, 'K'), temperatureK)

    def test_convert_array_from_si(self):
        data = np.array([temperatureK, temperatureK2])
        data_copy = np.copy(data)
        result = self.units.convert_array_from_si(data, 'F')
        self.assertIsNot(result, data)
        self.assertFalse(np.array_equal(result, data))
        self.assertTrue(np.array_equal(data, data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    def test_convert_array_from_si_no_conversion_required(self):
        data = np.array([temperatureK, temperatureK2])
        result = self.units.convert_array_from_si(data, 'K')
        self.assertIs(result, data)

    def test_convert_array_from_si_always_return_copy(self):
        data = np.array([temperatureK, temperatureK2])
        result = self.units.convert_array_from_si(data, 'K', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(np.array_equal(result, data))

    def test_convert_array_from_si_inplace(self):
        data = np.array([temperatureK, temperatureK2])
        result = self.units.convert_array_from_si(data, 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    def test_convert_series_from_si(self):
        data = pd.Series([temperatureK, temperatureK2])
        data_copy = data.copy()
        result = self.units.convert_series_from_si(data, 'F')
        self.assertIsNot(result, data)
        self.assertFalse(result.equals(data))
        self.assertTrue(data.equals(data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    def test_convert_series_from_si_no_conversion_required(self):
        data = pd.Series([temperatureK, temperatureK2])
        result = self.units.convert_series_from_si(data, 'K')
        self.assertIs(result, data)

    def test_convert_series_from_si_always_return_copy(self):
        data = pd.Series([temperatureK, temperatureK2])
        result = self.units.convert_series_from_si(data, 'K', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(result.equals(data))

    def test_convert_series_from_si_inplace(self):
        data = pd.Series([temperatureK, temperatureK2])
        result = self.units.convert_series_from_si(data, 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    # TO SI
    def test_convert_value_to_si(self):
        self.assertAlmostEqual(self.units.convert_value_to_si(temperatureC, 'C'), temperatureK, delta=0.01)
        self.assertAlmostEqual(self.units.convert_value_to_si(temperatureF, 'F'), temperatureK, delta=0.01)
        self.assertEqual(self.units.convert_value_to_si(3, 'e-3'), 0.003)
        self.assertEqual(self.units.convert_value_to_si(3, 'e-6'), 0.000003)

        self.assertAlmostEqual(self.units.convert_value_to_si(temperatureC, 'C', True), temperatureC, delta=0.01)
        self.assertAlmostEqual(self.units.convert_value_to_si(temperatureF, 'F', True), temperatureF * 5 / 9, delta=0.01)

        self.assertEqual(self.units.convert_value_to_si(temperatureK, 'K'), temperatureK)

    def test_convert_array_to_si(self):
        data = np.array([temperatureF, temperatureF2])
        data_copy = np.copy(data)
        result = self.units.convert_array_to_si(data, 'F')
        self.assertIsNot(result, data)
        self.assertFalse(np.array_equal(result, data))
        self.assertTrue(np.array_equal(data, data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureK, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureK2, delta=0.01)

    def test_convert_array_to_si_no_conversion_required(self):
        data = np.array([temperatureK, temperatureK2])
        result = self.units.convert_array_to_si(data, 'K')
        self.assertIs(result, data)

    def test_convert_array_to_si_always_return_copy(self):
        data = np.array([temperatureK, temperatureK2])
        result = self.units.convert_array_to_si(data, 'K', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(np.array_equal(result, data))

    def test_convert_array_to_si_inplace(self):
        data = np.array([temperatureF, temperatureF2])
        result = self.units.convert_array_to_si(data, 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureK, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureK2, delta=0.01)

    def test_convert_series_to_si(self):
        data = pd.Series([temperatureF, temperatureF2])
        data_copy = data.copy()
        result = self.units.convert_series_to_si(data, 'F')
        self.assertIsNot(result, data)
        self.assertFalse(result.equals(data))
        self.assertTrue(data.equals(data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureK, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureK2, delta=0.01)

    def test_convert_series_to_si_no_conversion_required(self):
        data = pd.Series([temperatureK, temperatureK2])
        result = self.units.convert_series_to_si(data, 'K')
        self.assertIs(result, data)

    def test_convert_series_to_si_always_return_copy(self):
        data = pd.Series([temperatureK, temperatureK2])
        result = self.units.convert_series_to_si(data, 'K', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(result.equals(data))

    def test_convert_series_to_si_inplace(self):
        data = pd.Series([temperatureF, temperatureF2])
        result = self.units.convert_series_to_si(data, 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureK, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureK2, delta=0.01)

    # BETWEEN UNITS
    def test_convert_value_between_units(self):
        self.assertAlmostEqual(self.units.convert_value_between_units(temperatureC, 'C', 'F'), temperatureF, delta=0.01)

    def test_convert_array_between_units(self):
        data = np.array([temperatureC, temperatureC2])
        data_copy = np.copy(data)
        result = self.units.convert_array_between_units(data, 'C', 'F')
        self.assertIsNot(result, data)
        self.assertFalse(np.array_equal(result, data))
        self.assertTrue(np.array_equal(data, data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

        # It should convert between non-si units when only factor
        data = np.array([1000.0, 2000.0])
        result = self.units.convert_array_between_units(data, 'mm', 'km')
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 0.001)
        self.assertEqual(result[1], 0.002)

    def test_convert_array_between_units_no_conversion_required(self):
        data = np.array([temperatureC, temperatureC2])
        result = self.units.convert_array_between_units(data, 'C', 'C')
        self.assertIs(result, data)

    def test_convert_array_between_units_always_return_copy(self):
        data = np.array([temperatureC, temperatureC2])
        result = self.units.convert_array_between_units(data, 'C', 'C', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(np.array_equal(result, data))

    def test_convert_array_between_units_inplace(self):
        data = np.array([temperatureC, temperatureC2])
        result = self.units.convert_array_between_units(data, 'C', 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    def test_convert_series_between_units(self):
        data = pd.Series([temperatureC, temperatureC2])
        data_copy = data.copy()
        result = self.units.convert_series_between_units(data, 'C', 'F')
        self.assertIsNot(result, data)
        self.assertFalse(result.equals(data))
        self.assertTrue(data.equals(data_copy))
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

        # It should convert between non-si units when only factor
        data = pd.Series([1000.0, 2000.0])
        result = self.units.convert_series_between_units(data, 'mm', 'km')
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 0.001)
        self.assertEqual(result[1], 0.002)

    def test_convert_series_between_units_no_conversion_required(self):
        data = pd.Series([temperatureC, temperatureC2])
        result = self.units.convert_series_between_units(data, 'C', 'C')
        self.assertIs(result, data)

    def test_convert_series_between_units_always_return_copy(self):
        data = pd.Series([temperatureC, temperatureC2])
        result = self.units.convert_series_between_units(data, 'C', 'C', always_return_copy=True)
        self.assertIsNot(result, data)
        self.assertTrue(result.equals(data))

    def test_convert_series_between_units_inplace(self):
        data = pd.Series([temperatureC, temperatureC2])
        result = self.units.convert_series_between_units(data, 'C', 'F', inplace=True)
        self.assertIs(result, data)
        self.assertEqual(len(result), 2)
        self.assertAlmostEqual(result[0], temperatureF, delta=0.01)
        self.assertAlmostEqual(result[1], temperatureF2, delta=0.01)

    def test_is_conversion_required(self):
        self.assertTrue(self.units.get_conversion_between_units('mm', 'km').is_conversion_required)
        self.assertTrue(self.units.get_conversion_between_units('m', 'km').is_conversion_required)
        self.assertTrue(self.units.get_conversion_between_units('km', 'm').is_conversion_required)
        self.assertTrue(self.units.get_conversion_between_units('C', 'F').is_conversion_required)
        self.assertFalse(self.units.get_conversion_between_units('F', 'F').is_conversion_required)
        self.assertFalse(self.units.get_conversion_between_units('m', 'm').is_conversion_required)
        self.assertFalse(self.units.get_conversion_between_units('mm', 'mm').is_conversion_required)



from typing import Optional

import canopy


async def delete_study(
        session: canopy.Session,
        study_id: str,
        tenant_id: Optional[str] = None,
        undelete: Optional[bool] = False) -> None:

    session.authentication.authenticate()

    if tenant_id is None:
        tenant_id = session.authentication.tenant_id

    study_api = canopy.openapi.StudyApi(session.async_client)
    await study_api.study_delete_study(
        tenant_id,
        study_id,
        **canopy.defined_kwargs(
            undelete=undelete))



import canopy
import unittest


class EnsureSimTypeStringTest(unittest.TestCase):

    def test_sim_type_string(self):
        self.assertEqual('ApexSim', canopy.ensure_sim_type_string('ApexSim'))

    def test_study_type_string(self):
        self.assertEqual('ApexSim', canopy.ensure_sim_type_string('apexSim'))

    def test_short_strings(self):
        self.assertEqual('', canopy.ensure_sim_type_string(''))
        self.assertEqual('A', canopy.ensure_sim_type_string('A'))
        self.assertEqual('A', canopy.ensure_sim_type_string('a'))
        self.assertEqual('Ap', canopy.ensure_sim_type_string('Ap'))
        self.assertEqual('Ap', canopy.ensure_sim_type_string('ap'))



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class GetTenantUsersQueryResultUserItem(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'user_id': 'object',
        'username': 'object',
        'email': 'object',
        'roles': 'list[str]',
        'is_enabled': 'bool'
    }

    attribute_map = {
        'user_id': 'userId',
        'username': 'username',
        'email': 'email',
        'roles': 'roles',
        'is_enabled': 'isEnabled'
    }

    def __init__(self, user_id=None, username=None, email=None, roles=None, is_enabled=None, local_vars_configuration=None):  # noqa: E501
        """GetTenantUsersQueryResultUserItem - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._user_id = None
        self._username = None
        self._email = None
        self._roles = None
        self._is_enabled = None
        self.discriminator = None

        self.user_id = user_id
        self.username = username
        self.email = email
        self.roles = roles
        self.is_enabled = is_enabled

    @property
    def user_id(self):
        """Gets the user_id of this GetTenantUsersQueryResultUserItem.  # noqa: E501


        :return: The user_id of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :rtype: object
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this GetTenantUsersQueryResultUserItem.


        :param user_id: The user_id of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :type user_id: object
        """
        if self.local_vars_configuration.client_side_validation and user_id is None:  # noqa: E501
            raise ValueError("Invalid value for `user_id`, must not be `None`")  # noqa: E501

        self._user_id = user_id

    @property
    def username(self):
        """Gets the username of this GetTenantUsersQueryResultUserItem.  # noqa: E501


        :return: The username of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :rtype: object
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this GetTenantUsersQueryResultUserItem.


        :param username: The username of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :type username: object
        """
        if self.local_vars_configuration.client_side_validation and username is None:  # noqa: E501
            raise ValueError("Invalid value for `username`, must not be `None`")  # noqa: E501

        self._username = username

    @property
    def email(self):
        """Gets the email of this GetTenantUsersQueryResultUserItem.  # noqa: E501


        :return: The email of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :rtype: object
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this GetTenantUsersQueryResultUserItem.


        :param email: The email of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :type email: object
        """
        if self.local_vars_configuration.client_side_validation and email is None:  # noqa: E501
            raise ValueError("Invalid value for `email`, must not be `None`")  # noqa: E501

        self._email = email

    @property
    def roles(self):
        """Gets the roles of this GetTenantUsersQueryResultUserItem.  # noqa: E501


        :return: The roles of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :rtype: list[str]
        """
        return self._roles

    @roles.setter
    def roles(self, roles):
        """Sets the roles of this GetTenantUsersQueryResultUserItem.


        :param roles: The roles of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :type roles: list[str]
        """
        if self.local_vars_configuration.client_side_validation and roles is None:  # noqa: E501
            raise ValueError("Invalid value for `roles`, must not be `None`")  # noqa: E501

        self._roles = roles

    @property
    def is_enabled(self):
        """Gets the is_enabled of this GetTenantUsersQueryResultUserItem.  # noqa: E501


        :return: The is_enabled of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :rtype: bool
        """
        return self._is_enabled

    @is_enabled.setter
    def is_enabled(self, is_enabled):
        """Sets the is_enabled of this GetTenantUsersQueryResultUserItem.


        :param is_enabled: The is_enabled of this GetTenantUsersQueryResultUserItem.  # noqa: E501
        :type is_enabled: bool
        """
        if self.local_vars_configuration.client_side_validation and is_enabled is None:  # noqa: E501
            raise ValueError("Invalid value for `is_enabled`, must not be `None`")  # noqa: E501

        self._is_enabled = is_enabled

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GetTenantUsersQueryResultUserItem):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GetTenantUsersQueryResultUserItem):
            return True

        return self.to_dict() != other.to_dict()



from typing import Any, Optional

import canopy
import json


class ConfigResult:
    _converted_data: Optional[Any] = None

    def __init__(self, document: canopy.openapi.CanopyDocument, user_information: canopy.openapi.DocumentUserInformation):
        self._document = document
        self._user_information = user_information

    @property
    def document(self) -> canopy.openapi.CanopyDocument:
        return self._document

    @property
    def config_id(self) -> str:
        return self._document.document_id

    @property
    def data(self) -> Any:
        if self._converted_data is None:
            self._converted_data = canopy.dict_to_object(self._get_raw_data())
        return self._converted_data

    @property
    def is_data_converted(self) -> bool:
        return self._converted_data is not None

    @property
    def raw_data(self) -> Any:
        return self._converted_data if self.is_data_converted else self.document.data

    @property
    def user_information(self) -> canopy.openapi.DocumentUserInformation:
        return self._user_information

    def to_local_config(self):
        return canopy.LocalConfig(
            self.document.sub_type,
            self.document.name,
            self.raw_data,
            properties=canopy.ensure_dict(self.document.properties),
            notes=self.document.notes,
            config_id=self.document.document_id,
            user_id=self.document.user_id,
            is_edited=False)

    def _get_raw_data(self):
        if self._document.sub_type == canopy.Constants.config_sub_tree_document_type:
            return self._document.data['definition']
        else:
            return self._document.data

    def to_dict(self):
        return {
            'config_id': self.config_id,
            'document': self._document.to_dict(),
            'user_information': self._user_information.to_dict(),
            'is_data_converted': self.is_data_converted,
            'data': self.raw_data
        }

    def __repr__(self):
        return 'canopy.ConfigResult(%r,%r)' % \
               (self._document, self._user_information)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2, default=str)



from typing import Optional, List, Mapping

import canopy


class UserSettingsCache(object):

    def __init__(self, client: canopy.openapi.ApiClient, authentication: canopy.Authentication):
        self._client = client
        self._authentication = authentication
        self._data: Optional[canopy.openapi.UserSettings] = None
        self._units: Optional[Mapping[str, str]] = None

    def reload(self):
        user_settings_api = canopy.openapi.UserSettingsApi(self._client)
        user_settings_result: canopy.openapi.GetUserSettingsQueryResult = user_settings_api.user_settings_get_user_settings(
            self._authentication.tenant_id,
            self._authentication.user_id)

        user_settings: canopy.openapi.UserSettings = user_settings_result.settings
        self._data = user_settings
        channels: List[canopy.openapi.ChannelSettings] = user_settings.channels
        self._units = {item.name: item.units for item in channels}

    def load_if_required(self):
        if self._data is None:
            self.reload()

    @property
    def data(self) -> canopy.openapi.UserSettings:
        if self._data is None:
            self.reload()

        if self._data is None:
            raise RuntimeError('User settings not found.')

        return self._data

    def get_channel_units(self, channel_name: str) -> Optional[str]:
        if self._units is None:
            self.reload()

        if self._units is None:
            raise RuntimeError('User settings not found.')

        if channel_name in self._units:
            return self._units[channel_name]

        return None



import canopy


class EnvironmentData(object):
    def __init__(
            self,
            authentication: canopy.AuthenticationData):
        self._authentication = authentication

    @property
    def authentication(self) -> canopy.AuthenticationData:
        return self._authentication

    def create_session(self) -> canopy.Session:
        return canopy.Session(self._authentication)
        # return canopy.Session(self._authentication, proxy=canopy.ProxyConfiguration('http://localhost:8888', '1', '1'))



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class GetPoolsItem(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'pool_id': 'object',
        'current_dedicated': 'int',
        'target_dedicated': 'int',
        'current_low_priority': 'int',
        'target_low_priority': 'int',
        'intervals': 'list[GetPoolsItemInterval]'
    }

    attribute_map = {
        'pool_id': 'poolId',
        'current_dedicated': 'currentDedicated',
        'target_dedicated': 'targetDedicated',
        'current_low_priority': 'currentLowPriority',
        'target_low_priority': 'targetLowPriority',
        'intervals': 'intervals'
    }

    def __init__(self, pool_id=None, current_dedicated=None, target_dedicated=None, current_low_priority=None, target_low_priority=None, intervals=None, local_vars_configuration=None):  # noqa: E501
        """GetPoolsItem - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._pool_id = None
        self._current_dedicated = None
        self._target_dedicated = None
        self._current_low_priority = None
        self._target_low_priority = None
        self._intervals = None
        self.discriminator = None

        self.pool_id = pool_id
        self.current_dedicated = current_dedicated
        self.target_dedicated = target_dedicated
        self.current_low_priority = current_low_priority
        self.target_low_priority = target_low_priority
        self.intervals = intervals

    @property
    def pool_id(self):
        """Gets the pool_id of this GetPoolsItem.  # noqa: E501


        :return: The pool_id of this GetPoolsItem.  # noqa: E501
        :rtype: object
        """
        return self._pool_id

    @pool_id.setter
    def pool_id(self, pool_id):
        """Sets the pool_id of this GetPoolsItem.


        :param pool_id: The pool_id of this GetPoolsItem.  # noqa: E501
        :type pool_id: object
        """
        if self.local_vars_configuration.client_side_validation and pool_id is None:  # noqa: E501
            raise ValueError("Invalid value for `pool_id`, must not be `None`")  # noqa: E501

        self._pool_id = pool_id

    @property
    def current_dedicated(self):
        """Gets the current_dedicated of this GetPoolsItem.  # noqa: E501


        :return: The current_dedicated of this GetPoolsItem.  # noqa: E501
        :rtype: int
        """
        return self._current_dedicated

    @current_dedicated.setter
    def current_dedicated(self, current_dedicated):
        """Sets the current_dedicated of this GetPoolsItem.


        :param current_dedicated: The current_dedicated of this GetPoolsItem.  # noqa: E501
        :type current_dedicated: int
        """
        if self.local_vars_configuration.client_side_validation and current_dedicated is None:  # noqa: E501
            raise ValueError("Invalid value for `current_dedicated`, must not be `None`")  # noqa: E501

        self._current_dedicated = current_dedicated

    @property
    def target_dedicated(self):
        """Gets the target_dedicated of this GetPoolsItem.  # noqa: E501


        :return: The target_dedicated of this GetPoolsItem.  # noqa: E501
        :rtype: int
        """
        return self._target_dedicated

    @target_dedicated.setter
    def target_dedicated(self, target_dedicated):
        """Sets the target_dedicated of this GetPoolsItem.


        :param target_dedicated: The target_dedicated of this GetPoolsItem.  # noqa: E501
        :type target_dedicated: int
        """
        if self.local_vars_configuration.client_side_validation and target_dedicated is None:  # noqa: E501
            raise ValueError("Invalid value for `target_dedicated`, must not be `None`")  # noqa: E501

        self._target_dedicated = target_dedicated

    @property
    def current_low_priority(self):
        """Gets the current_low_priority of this GetPoolsItem.  # noqa: E501


        :return: The current_low_priority of this GetPoolsItem.  # noqa: E501
        :rtype: int
        """
        return self._current_low_priority

    @current_low_priority.setter
    def current_low_priority(self, current_low_priority):
        """Sets the current_low_priority of this GetPoolsItem.


        :param current_low_priority: The current_low_priority of this GetPoolsItem.  # noqa: E501
        :type current_low_priority: int
        """
        if self.local_vars_configuration.client_side_validation and current_low_priority is None:  # noqa: E501
            raise ValueError("Invalid value for `current_low_priority`, must not be `None`")  # noqa: E501

        self._current_low_priority = current_low_priority

    @property
    def target_low_priority(self):
        """Gets the target_low_priority of this GetPoolsItem.  # noqa: E501


        :return: The target_low_priority of this GetPoolsItem.  # noqa: E501
        :rtype: int
        """
        return self._target_low_priority

    @target_low_priority.setter
    def target_low_priority(self, target_low_priority):
        """Sets the target_low_priority of this GetPoolsItem.


        :param target_low_priority: The target_low_priority of this GetPoolsItem.  # noqa: E501
        :type target_low_priority: int
        """
        if self.local_vars_configuration.client_side_validation and target_low_priority is None:  # noqa: E501
            raise ValueError("Invalid value for `target_low_priority`, must not be `None`")  # noqa: E501

        self._target_low_priority = target_low_priority

    @property
    def intervals(self):
        """Gets the intervals of this GetPoolsItem.  # noqa: E501


        :return: The intervals of this GetPoolsItem.  # noqa: E501
        :rtype: list[GetPoolsItemInterval]
        """
        return self._intervals

    @intervals.setter
    def intervals(self, intervals):
        """Sets the intervals of this GetPoolsItem.


        :param intervals: The intervals of this GetPoolsItem.  # noqa: E501
        :type intervals: list[GetPoolsItemInterval]
        """
        if self.local_vars_configuration.client_side_validation and intervals is None:  # noqa: E501
            raise ValueError("Invalid value for `intervals`, must not be `None`")  # noqa: E501

        self._intervals = intervals

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GetPoolsItem):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GetPoolsItem):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class StudyDocument(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'error_messages': 'list[str]',
        'job_count': 'int',
        'dispatched_job_count': 'int',
        'completed_job_count': 'int',
        'succeeded_job_count': 'int',
        'dispatched_compute_credits': 'float',
        'completed_compute_credits': 'float',
        'succeeded_compute_credits': 'float',
        'dispatched_storage_credits': 'float',
        'completed_storage_credits': 'float',
        'succeeded_storage_credits': 'float',
        'succeeded_simulation_count': 'int',
        'seed': 'int',
        'is_transient': 'bool',
        'execution_time_seconds': 'float',
        'study_type': 'str',
        'study_state': 'StudyState',
        'sources': 'list[StudyDocumentStudyDocumentDataSource]',
        'definition': 'object',
        'shard_dispatched_job_counts': 'list[int]',
        'shard_dispatched_compute_credits': 'list[float]',
        'shard_dispatched_storage_credits': 'list[float]'
    }

    attribute_map = {
        'error_messages': 'errorMessages',
        'job_count': 'jobCount',
        'dispatched_job_count': 'dispatchedJobCount',
        'completed_job_count': 'completedJobCount',
        'succeeded_job_count': 'succeededJobCount',
        'dispatched_compute_credits': 'dispatchedComputeCredits',
        'completed_compute_credits': 'completedComputeCredits',
        'succeeded_compute_credits': 'succeededComputeCredits',
        'dispatched_storage_credits': 'dispatchedStorageCredits',
        'completed_storage_credits': 'completedStorageCredits',
        'succeeded_storage_credits': 'succeededStorageCredits',
        'succeeded_simulation_count': 'succeededSimulationCount',
        'seed': 'seed',
        'is_transient': 'isTransient',
        'execution_time_seconds': 'executionTimeSeconds',
        'study_type': 'studyType',
        'study_state': 'studyState',
        'sources': 'sources',
        'definition': 'definition',
        'shard_dispatched_job_counts': 'shardDispatchedJobCounts',
        'shard_dispatched_compute_credits': 'shardDispatchedComputeCredits',
        'shard_dispatched_storage_credits': 'shardDispatchedStorageCredits'
    }

    def __init__(self, error_messages=None, job_count=None, dispatched_job_count=None, completed_job_count=None, succeeded_job_count=None, dispatched_compute_credits=None, completed_compute_credits=None, succeeded_compute_credits=None, dispatched_storage_credits=None, completed_storage_credits=None, succeeded_storage_credits=None, succeeded_simulation_count=None, seed=None, is_transient=None, execution_time_seconds=None, study_type=None, study_state=None, sources=None, definition=None, shard_dispatched_job_counts=None, shard_dispatched_compute_credits=None, shard_dispatched_storage_credits=None, local_vars_configuration=None):  # noqa: E501
        """StudyDocument - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._error_messages = None
        self._job_count = None
        self._dispatched_job_count = None
        self._completed_job_count = None
        self._succeeded_job_count = None
        self._dispatched_compute_credits = None
        self._completed_compute_credits = None
        self._succeeded_compute_credits = None
        self._dispatched_storage_credits = None
        self._completed_storage_credits = None
        self._succeeded_storage_credits = None
        self._succeeded_simulation_count = None
        self._seed = None
        self._is_transient = None
        self._execution_time_seconds = None
        self._study_type = None
        self._study_state = None
        self._sources = None
        self._definition = None
        self._shard_dispatched_job_counts = None
        self._shard_dispatched_compute_credits = None
        self._shard_dispatched_storage_credits = None
        self.discriminator = None

        self.error_messages = error_messages
        self.job_count = job_count
        self.dispatched_job_count = dispatched_job_count
        self.completed_job_count = completed_job_count
        self.succeeded_job_count = succeeded_job_count
        if dispatched_compute_credits is not None:
            self.dispatched_compute_credits = dispatched_compute_credits
        if completed_compute_credits is not None:
            self.completed_compute_credits = completed_compute_credits
        self.succeeded_compute_credits = succeeded_compute_credits
        if dispatched_storage_credits is not None:
            self.dispatched_storage_credits = dispatched_storage_credits
        if completed_storage_credits is not None:
            self.completed_storage_credits = completed_storage_credits
        self.succeeded_storage_credits = succeeded_storage_credits
        if succeeded_simulation_count is not None:
            self.succeeded_simulation_count = succeeded_simulation_count
        self.seed = seed
        self.is_transient = is_transient
        self.execution_time_seconds = execution_time_seconds
        self.study_type = study_type
        self.study_state = study_state
        self.sources = sources
        self.definition = definition
        self.shard_dispatched_job_counts = shard_dispatched_job_counts
        self.shard_dispatched_compute_credits = shard_dispatched_compute_credits
        self.shard_dispatched_storage_credits = shard_dispatched_storage_credits

    @property
    def error_messages(self):
        """Gets the error_messages of this StudyDocument.  # noqa: E501


        :return: The error_messages of this StudyDocument.  # noqa: E501
        :rtype: list[str]
        """
        return self._error_messages

    @error_messages.setter
    def error_messages(self, error_messages):
        """Sets the error_messages of this StudyDocument.


        :param error_messages: The error_messages of this StudyDocument.  # noqa: E501
        :type error_messages: list[str]
        """
        if self.local_vars_configuration.client_side_validation and error_messages is None:  # noqa: E501
            raise ValueError("Invalid value for `error_messages`, must not be `None`")  # noqa: E501

        self._error_messages = error_messages

    @property
    def job_count(self):
        """Gets the job_count of this StudyDocument.  # noqa: E501


        :return: The job_count of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._job_count

    @job_count.setter
    def job_count(self, job_count):
        """Sets the job_count of this StudyDocument.


        :param job_count: The job_count of this StudyDocument.  # noqa: E501
        :type job_count: int
        """
        if self.local_vars_configuration.client_side_validation and job_count is None:  # noqa: E501
            raise ValueError("Invalid value for `job_count`, must not be `None`")  # noqa: E501

        self._job_count = job_count

    @property
    def dispatched_job_count(self):
        """Gets the dispatched_job_count of this StudyDocument.  # noqa: E501


        :return: The dispatched_job_count of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._dispatched_job_count

    @dispatched_job_count.setter
    def dispatched_job_count(self, dispatched_job_count):
        """Sets the dispatched_job_count of this StudyDocument.


        :param dispatched_job_count: The dispatched_job_count of this StudyDocument.  # noqa: E501
        :type dispatched_job_count: int
        """
        if self.local_vars_configuration.client_side_validation and dispatched_job_count is None:  # noqa: E501
            raise ValueError("Invalid value for `dispatched_job_count`, must not be `None`")  # noqa: E501

        self._dispatched_job_count = dispatched_job_count

    @property
    def completed_job_count(self):
        """Gets the completed_job_count of this StudyDocument.  # noqa: E501


        :return: The completed_job_count of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._completed_job_count

    @completed_job_count.setter
    def completed_job_count(self, completed_job_count):
        """Sets the completed_job_count of this StudyDocument.


        :param completed_job_count: The completed_job_count of this StudyDocument.  # noqa: E501
        :type completed_job_count: int
        """
        if self.local_vars_configuration.client_side_validation and completed_job_count is None:  # noqa: E501
            raise ValueError("Invalid value for `completed_job_count`, must not be `None`")  # noqa: E501

        self._completed_job_count = completed_job_count

    @property
    def succeeded_job_count(self):
        """Gets the succeeded_job_count of this StudyDocument.  # noqa: E501


        :return: The succeeded_job_count of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._succeeded_job_count

    @succeeded_job_count.setter
    def succeeded_job_count(self, succeeded_job_count):
        """Sets the succeeded_job_count of this StudyDocument.


        :param succeeded_job_count: The succeeded_job_count of this StudyDocument.  # noqa: E501
        :type succeeded_job_count: int
        """
        if self.local_vars_configuration.client_side_validation and succeeded_job_count is None:  # noqa: E501
            raise ValueError("Invalid value for `succeeded_job_count`, must not be `None`")  # noqa: E501

        self._succeeded_job_count = succeeded_job_count

    @property
    def dispatched_compute_credits(self):
        """Gets the dispatched_compute_credits of this StudyDocument.  # noqa: E501


        :return: The dispatched_compute_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._dispatched_compute_credits

    @dispatched_compute_credits.setter
    def dispatched_compute_credits(self, dispatched_compute_credits):
        """Sets the dispatched_compute_credits of this StudyDocument.


        :param dispatched_compute_credits: The dispatched_compute_credits of this StudyDocument.  # noqa: E501
        :type dispatched_compute_credits: float
        """

        self._dispatched_compute_credits = dispatched_compute_credits

    @property
    def completed_compute_credits(self):
        """Gets the completed_compute_credits of this StudyDocument.  # noqa: E501


        :return: The completed_compute_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._completed_compute_credits

    @completed_compute_credits.setter
    def completed_compute_credits(self, completed_compute_credits):
        """Sets the completed_compute_credits of this StudyDocument.


        :param completed_compute_credits: The completed_compute_credits of this StudyDocument.  # noqa: E501
        :type completed_compute_credits: float
        """

        self._completed_compute_credits = completed_compute_credits

    @property
    def succeeded_compute_credits(self):
        """Gets the succeeded_compute_credits of this StudyDocument.  # noqa: E501


        :return: The succeeded_compute_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._succeeded_compute_credits

    @succeeded_compute_credits.setter
    def succeeded_compute_credits(self, succeeded_compute_credits):
        """Sets the succeeded_compute_credits of this StudyDocument.


        :param succeeded_compute_credits: The succeeded_compute_credits of this StudyDocument.  # noqa: E501
        :type succeeded_compute_credits: float
        """
        if self.local_vars_configuration.client_side_validation and succeeded_compute_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `succeeded_compute_credits`, must not be `None`")  # noqa: E501

        self._succeeded_compute_credits = succeeded_compute_credits

    @property
    def dispatched_storage_credits(self):
        """Gets the dispatched_storage_credits of this StudyDocument.  # noqa: E501


        :return: The dispatched_storage_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._dispatched_storage_credits

    @dispatched_storage_credits.setter
    def dispatched_storage_credits(self, dispatched_storage_credits):
        """Sets the dispatched_storage_credits of this StudyDocument.


        :param dispatched_storage_credits: The dispatched_storage_credits of this StudyDocument.  # noqa: E501
        :type dispatched_storage_credits: float
        """

        self._dispatched_storage_credits = dispatched_storage_credits

    @property
    def completed_storage_credits(self):
        """Gets the completed_storage_credits of this StudyDocument.  # noqa: E501


        :return: The completed_storage_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._completed_storage_credits

    @completed_storage_credits.setter
    def completed_storage_credits(self, completed_storage_credits):
        """Sets the completed_storage_credits of this StudyDocument.


        :param completed_storage_credits: The completed_storage_credits of this StudyDocument.  # noqa: E501
        :type completed_storage_credits: float
        """

        self._completed_storage_credits = completed_storage_credits

    @property
    def succeeded_storage_credits(self):
        """Gets the succeeded_storage_credits of this StudyDocument.  # noqa: E501


        :return: The succeeded_storage_credits of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._succeeded_storage_credits

    @succeeded_storage_credits.setter
    def succeeded_storage_credits(self, succeeded_storage_credits):
        """Sets the succeeded_storage_credits of this StudyDocument.


        :param succeeded_storage_credits: The succeeded_storage_credits of this StudyDocument.  # noqa: E501
        :type succeeded_storage_credits: float
        """
        if self.local_vars_configuration.client_side_validation and succeeded_storage_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `succeeded_storage_credits`, must not be `None`")  # noqa: E501

        self._succeeded_storage_credits = succeeded_storage_credits

    @property
    def succeeded_simulation_count(self):
        """Gets the succeeded_simulation_count of this StudyDocument.  # noqa: E501


        :return: The succeeded_simulation_count of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._succeeded_simulation_count

    @succeeded_simulation_count.setter
    def succeeded_simulation_count(self, succeeded_simulation_count):
        """Sets the succeeded_simulation_count of this StudyDocument.


        :param succeeded_simulation_count: The succeeded_simulation_count of this StudyDocument.  # noqa: E501
        :type succeeded_simulation_count: int
        """

        self._succeeded_simulation_count = succeeded_simulation_count

    @property
    def seed(self):
        """Gets the seed of this StudyDocument.  # noqa: E501


        :return: The seed of this StudyDocument.  # noqa: E501
        :rtype: int
        """
        return self._seed

    @seed.setter
    def seed(self, seed):
        """Sets the seed of this StudyDocument.


        :param seed: The seed of this StudyDocument.  # noqa: E501
        :type seed: int
        """
        if self.local_vars_configuration.client_side_validation and seed is None:  # noqa: E501
            raise ValueError("Invalid value for `seed`, must not be `None`")  # noqa: E501

        self._seed = seed

    @property
    def is_transient(self):
        """Gets the is_transient of this StudyDocument.  # noqa: E501


        :return: The is_transient of this StudyDocument.  # noqa: E501
        :rtype: bool
        """
        return self._is_transient

    @is_transient.setter
    def is_transient(self, is_transient):
        """Sets the is_transient of this StudyDocument.


        :param is_transient: The is_transient of this StudyDocument.  # noqa: E501
        :type is_transient: bool
        """
        if self.local_vars_configuration.client_side_validation and is_transient is None:  # noqa: E501
            raise ValueError("Invalid value for `is_transient`, must not be `None`")  # noqa: E501

        self._is_transient = is_transient

    @property
    def execution_time_seconds(self):
        """Gets the execution_time_seconds of this StudyDocument.  # noqa: E501


        :return: The execution_time_seconds of this StudyDocument.  # noqa: E501
        :rtype: float
        """
        return self._execution_time_seconds

    @execution_time_seconds.setter
    def execution_time_seconds(self, execution_time_seconds):
        """Sets the execution_time_seconds of this StudyDocument.


        :param execution_time_seconds: The execution_time_seconds of this StudyDocument.  # noqa: E501
        :type execution_time_seconds: float
        """
        if self.local_vars_configuration.client_side_validation and execution_time_seconds is None:  # noqa: E501
            raise ValueError("Invalid value for `execution_time_seconds`, must not be `None`")  # noqa: E501

        self._execution_time_seconds = execution_time_seconds

    @property
    def study_type(self):
        """Gets the study_type of this StudyDocument.  # noqa: E501


        :return: The study_type of this StudyDocument.  # noqa: E501
        :rtype: str
        """
        return self._study_type

    @study_type.setter
    def study_type(self, study_type):
        """Sets the study_type of this StudyDocument.


        :param study_type: The study_type of this StudyDocument.  # noqa: E501
        :type study_type: str
        """
        if self.local_vars_configuration.client_side_validation and study_type is None:  # noqa: E501
            raise ValueError("Invalid value for `study_type`, must not be `None`")  # noqa: E501

        self._study_type = study_type

    @property
    def study_state(self):
        """Gets the study_state of this StudyDocument.  # noqa: E501


        :return: The study_state of this StudyDocument.  # noqa: E501
        :rtype: StudyState
        """
        return self._study_state

    @study_state.setter
    def study_state(self, study_state):
        """Sets the study_state of this StudyDocument.


        :param study_state: The study_state of this StudyDocument.  # noqa: E501
        :type study_state: StudyState
        """

        self._study_state = study_state

    @property
    def sources(self):
        """Gets the sources of this StudyDocument.  # noqa: E501


        :return: The sources of this StudyDocument.  # noqa: E501
        :rtype: list[StudyDocumentStudyDocumentDataSource]
        """
        return self._sources

    @sources.setter
    def sources(self, sources):
        """Sets the sources of this StudyDocument.


        :param sources: The sources of this StudyDocument.  # noqa: E501
        :type sources: list[StudyDocumentStudyDocumentDataSource]
        """
        if self.local_vars_configuration.client_side_validation and sources is None:  # noqa: E501
            raise ValueError("Invalid value for `sources`, must not be `None`")  # noqa: E501

        self._sources = sources

    @property
    def definition(self):
        """Gets the definition of this StudyDocument.  # noqa: E501


        :return: The definition of this StudyDocument.  # noqa: E501
        :rtype: object
        """
        return self._definition

    @definition.setter
    def definition(self, definition):
        """Sets the definition of this StudyDocument.


        :param definition: The definition of this StudyDocument.  # noqa: E501
        :type definition: object
        """

        self._definition = definition

    @property
    def shard_dispatched_job_counts(self):
        """Gets the shard_dispatched_job_counts of this StudyDocument.  # noqa: E501


        :return: The shard_dispatched_job_counts of this StudyDocument.  # noqa: E501
        :rtype: list[int]
        """
        return self._shard_dispatched_job_counts

    @shard_dispatched_job_counts.setter
    def shard_dispatched_job_counts(self, shard_dispatched_job_counts):
        """Sets the shard_dispatched_job_counts of this StudyDocument.


        :param shard_dispatched_job_counts: The shard_dispatched_job_counts of this StudyDocument.  # noqa: E501
        :type shard_dispatched_job_counts: list[int]
        """

        self._shard_dispatched_job_counts = shard_dispatched_job_counts

    @property
    def shard_dispatched_compute_credits(self):
        """Gets the shard_dispatched_compute_credits of this StudyDocument.  # noqa: E501


        :return: The shard_dispatched_compute_credits of this StudyDocument.  # noqa: E501
        :rtype: list[float]
        """
        return self._shard_dispatched_compute_credits

    @shard_dispatched_compute_credits.setter
    def shard_dispatched_compute_credits(self, shard_dispatched_compute_credits):
        """Sets the shard_dispatched_compute_credits of this StudyDocument.


        :param shard_dispatched_compute_credits: The shard_dispatched_compute_credits of this StudyDocument.  # noqa: E501
        :type shard_dispatched_compute_credits: list[float]
        """

        self._shard_dispatched_compute_credits = shard_dispatched_compute_credits

    @property
    def shard_dispatched_storage_credits(self):
        """Gets the shard_dispatched_storage_credits of this StudyDocument.  # noqa: E501


        :return: The shard_dispatched_storage_credits of this StudyDocument.  # noqa: E501
        :rtype: list[float]
        """
        return self._shard_dispatched_storage_credits

    @shard_dispatched_storage_credits.setter
    def shard_dispatched_storage_credits(self, shard_dispatched_storage_credits):
        """Sets the shard_dispatched_storage_credits of this StudyDocument.


        :param shard_dispatched_storage_credits: The shard_dispatched_storage_credits of this StudyDocument.  # noqa: E501
        :type shard_dispatched_storage_credits: list[float]
        """

        self._shard_dispatched_storage_credits = shard_dispatched_storage_credits

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StudyDocument):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, StudyDocument):
            return True

        return self.to_dict() != other.to_dict()



import asyncio
from typing import TypeVar, Callable, Awaitable
import logging
from aiohttp.client_exceptions import ClientResponseError, ClientConnectionError, ServerTimeoutError, \
    ServerConnectionError

logger = logging.getLogger(__name__)

TResult = TypeVar('TResult')      # Declare type variable


async def request_with_retry(call: Callable[[], Awaitable[TResult]], error_subject: str, throw_error: bool) -> TResult:
    try:
        try:
            return await call()
        except (asyncio.TimeoutError, ClientConnectionError, ServerConnectionError):
            # We'll retry once.
            logger.warning(f'Retrying {error_subject}')
            return await call()
    except (asyncio.TimeoutError, ServerTimeoutError):
        logger.warning(f'Timed out loading {error_subject}')
        if throw_error:
            raise
    except ClientConnectionError:
        logger.warning(f'Connection error loading {error_subject}')
        if throw_error:
            raise
    except ClientResponseError as e:
        logger.warning(f'Response error loading {error_subject} ({e.message})')
        if throw_error:
            raise

    return None



# coding: utf-8

"""
    Canopy.Api

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from canopy.openapi.configuration import Configuration


class GrantTypeHandlerResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'expires': 'str',
        'issued': 'str',
        'access_token': 'str',
        'canopysimulationsclient_id': 'str',
        'expires_in': 'int',
        'refresh_token': 'str',
        'roles': 'str',
        'tenant_id': 'str',
        'token_type': 'str',
        'user_id': 'str',
        'username': 'str'
    }

    attribute_map = {
        'expires': '.expires',
        'issued': '.issued',
        'access_token': 'access_token',
        'canopysimulationsclient_id': 'canopysimulations:client_id:',
        'expires_in': 'expires_in',
        'refresh_token': 'refresh_token',
        'roles': 'roles',
        'tenant_id': 'tenant_id',
        'token_type': 'token_type',
        'user_id': 'user_id',
        'username': 'username'
    }

    def __init__(self, expires=None, issued=None, access_token=None, canopysimulationsclient_id=None, expires_in=None, refresh_token=None, roles=None, tenant_id=None, token_type=None, user_id=None, username=None, local_vars_configuration=None):  # noqa: E501
        """GrantTypeHandlerResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._expires = None
        self._issued = None
        self._access_token = None
        self._canopysimulationsclient_id = None
        self._expires_in = None
        self._refresh_token = None
        self._roles = None
        self._tenant_id = None
        self._token_type = None
        self._user_id = None
        self._username = None
        self.discriminator = None

        self.expires = expires
        self.issued = issued
        self.access_token = access_token
        self.canopysimulationsclient_id = canopysimulationsclient_id
        if expires_in is not None:
            self.expires_in = expires_in
        self.refresh_token = refresh_token
        self.roles = roles
        self.tenant_id = tenant_id
        self.token_type = token_type
        self.user_id = user_id
        self.username = username

    @property
    def expires(self):
        """Gets the expires of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The expires of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._expires

    @expires.setter
    def expires(self, expires):
        """Sets the expires of this GrantTypeHandlerResponse.


        :param expires: The expires of this GrantTypeHandlerResponse.  # noqa: E501
        :type expires: str
        """

        self._expires = expires

    @property
    def issued(self):
        """Gets the issued of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The issued of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._issued

    @issued.setter
    def issued(self, issued):
        """Sets the issued of this GrantTypeHandlerResponse.


        :param issued: The issued of this GrantTypeHandlerResponse.  # noqa: E501
        :type issued: str
        """

        self._issued = issued

    @property
    def access_token(self):
        """Gets the access_token of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The access_token of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._access_token

    @access_token.setter
    def access_token(self, access_token):
        """Sets the access_token of this GrantTypeHandlerResponse.


        :param access_token: The access_token of this GrantTypeHandlerResponse.  # noqa: E501
        :type access_token: str
        """

        self._access_token = access_token

    @property
    def canopysimulationsclient_id(self):
        """Gets the canopysimulationsclient_id of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The canopysimulationsclient_id of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._canopysimulationsclient_id

    @canopysimulationsclient_id.setter
    def canopysimulationsclient_id(self, canopysimulationsclient_id):
        """Sets the canopysimulationsclient_id of this GrantTypeHandlerResponse.


        :param canopysimulationsclient_id: The canopysimulationsclient_id of this GrantTypeHandlerResponse.  # noqa: E501
        :type canopysimulationsclient_id: str
        """

        self._canopysimulationsclient_id = canopysimulationsclient_id

    @property
    def expires_in(self):
        """Gets the expires_in of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The expires_in of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: int
        """
        return self._expires_in

    @expires_in.setter
    def expires_in(self, expires_in):
        """Sets the expires_in of this GrantTypeHandlerResponse.


        :param expires_in: The expires_in of this GrantTypeHandlerResponse.  # noqa: E501
        :type expires_in: int
        """

        self._expires_in = expires_in

    @property
    def refresh_token(self):
        """Gets the refresh_token of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The refresh_token of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, refresh_token):
        """Sets the refresh_token of this GrantTypeHandlerResponse.


        :param refresh_token: The refresh_token of this GrantTypeHandlerResponse.  # noqa: E501
        :type refresh_token: str
        """

        self._refresh_token = refresh_token

    @property
    def roles(self):
        """Gets the roles of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The roles of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._roles

    @roles.setter
    def roles(self, roles):
        """Sets the roles of this GrantTypeHandlerResponse.


        :param roles: The roles of this GrantTypeHandlerResponse.  # noqa: E501
        :type roles: str
        """

        self._roles = roles

    @property
    def tenant_id(self):
        """Gets the tenant_id of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The tenant_id of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._tenant_id

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        """Sets the tenant_id of this GrantTypeHandlerResponse.


        :param tenant_id: The tenant_id of this GrantTypeHandlerResponse.  # noqa: E501
        :type tenant_id: str
        """

        self._tenant_id = tenant_id

    @property
    def token_type(self):
        """Gets the token_type of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The token_type of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._token_type

    @token_type.setter
    def token_type(self, token_type):
        """Sets the token_type of this GrantTypeHandlerResponse.


        :param token_type: The token_type of this GrantTypeHandlerResponse.  # noqa: E501
        :type token_type: str
        """

        self._token_type = token_type

    @property
    def user_id(self):
        """Gets the user_id of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The user_id of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this GrantTypeHandlerResponse.


        :param user_id: The user_id of this GrantTypeHandlerResponse.  # noqa: E501
        :type user_id: str
        """

        self._user_id = user_id

    @property
    def username(self):
        """Gets the username of this GrantTypeHandlerResponse.  # noqa: E501


        :return: The username of this GrantTypeHandlerResponse.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this GrantTypeHandlerResponse.


        :param username: The username of this GrantTypeHandlerResponse.  # noqa: E501
        :type username: str
        """

        self._username = username

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GrantTypeHandlerResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GrantTypeHandlerResponse):
            return True

        return self.to_dict() != other.to_dict()



from typing import TypeVar, Generic
import canopy
import json

T = TypeVar('T')


class SerializableValue(Generic[T]):
    def __init__(
            self,
            session: canopy.Session,
            value: T):
        self._session = session
        self.value = value

    def serialize(self) -> str:
        return json.dumps(self._session.async_client.sanitize_for_serialization(self.value))
